import argparse
import gc
import os
import random
import shutil
import time
from copy import deepcopy
from functools import cached_property
from multiprocessing import Pool, cpu_count
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Iterable
from warnings import warn

import psutil
from frozendict import frozendict
from tqdm import tqdm

from bgpy.as_graphs.base import ASGraph, ASGraphConstructor
from bgpy.as_graphs.caida_as_graph import CAIDAASGraphConstructor
from bgpy.shared.constants import DIRS, SINGLE_DAY_CACHE_DIR, bgpy_logger
from bgpy.shared.enums import SpecialPercentAdoptions
from bgpy.simulation_engine import (
    BGP,
    ROV,
    BaseSimulationEngine,
    BGPFull,
    SimulationEngine,
)

from .as_graph_analyzers import ASGraphAnalyzer, BaseASGraphAnalyzer
from .graph_data_aggregator import GraphCategory, GraphDataAggregator
from .graphing import GraphFactory
from .scenarios import Scenario, ScenarioConfig, SubprefixHijack
from .utils import get_all_graph_categories

if TYPE_CHECKING:
    from multiprocessing.pool import ApplyResult

parser = argparse.ArgumentParser(description="Runs BGPy simulations")
parser.add_argument(
    "--num_trials",
    "--trials",
    dest="trials",
    type=int,
    default=1,
    help="Number of trials to run",
)
# parse known args to avoid crashing during pytest
args, _unknown = parser.parse_known_args()


class Simulation:
    """Runs simulations for BGP attack/defend scenarios"""

    def __init__(
        self,
        *,
        sim_name: str | None = None,
        percent_adoptions: tuple[float | SpecialPercentAdoptions, ...] = (
            SpecialPercentAdoptions.ONLY_ONE,
            0.1,
            0.2,
            0.5,
            0.8,
            0.99,
        ),
        scenario_configs: tuple[ScenarioConfig, ...] = (
            ScenarioConfig(
                ScenarioCls=SubprefixHijack,
                AdoptPolicyCls=ROV,
                BasePolicyCls=BGP,
            ),
        ),
        num_trials: int = args.trials,
        output_dir: Path | None = None,
        parse_cpus: int = max(cpu_count() - 1, 1),
        python_hash_seed: int | None = None,
        ASGraphConstructorCls: type[ASGraphConstructor] = CAIDAASGraphConstructor,
        as_graph_constructor_kwargs=frozendict(
            {
                "as_graph_collector_kwargs": frozendict(
                    {
                        # dl_time: datetime(),
                        "cache_dir": SINGLE_DAY_CACHE_DIR,
                    }
                ),
                "as_graph_kwargs": frozendict(
                    {
                        # When no ASNs are stored, .9gb/core
                        # When one set of cones is stored, 1.6gb/core
                        # When both sets of cones are stored, 2.3gb/core
                        "store_customer_cone_size": True,
                        "store_customer_cone_asns": False,
                        "store_provider_cone_size": False,
                        "store_provider_cone_asns": False,
                    }
                ),
                "tsv_path": None,  # Path.home() / "Desktop" / "caida.tsv",
            }
        ),
        SimulationEngineCls: type[BaseSimulationEngine] = SimulationEngine,
        ASGraphAnalyzerCls: type[BaseASGraphAnalyzer] = ASGraphAnalyzer,
        GraphDataAggregatorCls: type[GraphDataAggregator] = GraphDataAggregator,
        # Data plane tracking for traceback and GraphDataAggregatorCls
        data_plane_tracking: bool = True,
        # Control plane trackign for traceback and GraphDataAggregatorCls
        control_plane_tracking: bool = False,
        graph_categories: tuple[GraphCategory, ...] = tuple(get_all_graph_categories()),
    ) -> None:
        """Downloads relationship data, runs simulation

        Graphs -> A list of graph classes
        graph_path: Where to store the graphs. Should be a .tar.gz file
        assert_pypy: Ensures you are using pypy if true
        mp_method: Multiprocessing method
        """

        self.percent_adoptions: tuple[float | SpecialPercentAdoptions, ...] = (
            percent_adoptions
        )
        self.num_trials: int = num_trials
        self.parse_cpus: int = parse_cpus
        self.scenario_configs: tuple[ScenarioConfig, ...] = (
            self._get_filtered_scenario_configs(scenario_configs)
        )

        self.python_hash_seed: int | None = python_hash_seed
        self._seed_random()

        self.sim_name: str = sim_name if sim_name else self.default_sim_name
        self.output_dir: Path = output_dir if output_dir else self.default_output_dir
        self.output_dir.mkdir(parents=True, exist_ok=True)

        # Done here so that the caida files are cached
        # So that multiprocessing doesn't interfere with one another
        self.ASGraphConstructorCls: type[ASGraphConstructor] = ASGraphConstructorCls
        self.as_graph_constructor_kwargs = as_graph_constructor_kwargs
        # Validate before building the graph to give the user a few seconds to kill
        # the program if desired (if RAM would run out) or validation errors before
        # multi-second as graph loading takes place
        self._validate_init()

        self.SimulationEngineCls: type[BaseSimulationEngine] = SimulationEngineCls

        self.ASGraphAnalyzerCls: type[BaseASGraphAnalyzer] = ASGraphAnalyzerCls
        self.GraphDataAggregatorCls: type[GraphDataAggregator] = GraphDataAggregatorCls

        self.data_plane_tracking: bool = data_plane_tracking
        self.control_plane_tracking: bool = control_plane_tracking

        self.graph_categories: tuple[GraphCategory, ...] = graph_categories

        # Can't delete this since it gets deleted in multiprocessing for some reason
        # NOTE: Once pypy gets to 3.12, just pass delete=False to this
        with TemporaryDirectory() as tmp_dir:
            tmp_dir_str = tmp_dir
        self._tqdm_tracking_dir: Path = Path(tmp_dir_str)
        self._tqdm_tracking_dir.mkdir(parents=True)

    @property
    def default_sim_name(self) -> str:
        return "bgpy_sims"

    @property
    def default_output_dir(self) -> Path:
        return Path(DIRS.user_desktop_dir) / "sims" / self.sim_name

    def _get_filtered_scenario_configs(
        self, scenario_configs: tuple[ScenarioConfig, ...]
    ) -> tuple[ScenarioConfig, ...]:
        """Hook func to filter out inapplicable ScenarioConfigs"""

        return scenario_configs

    def _validate_init(self):
        """Validates inputs to __init__

        Specifically checks for:
        1. scenario config mismatch between adopting and base policies
        2. duplicate scenario labels
        3. RAM constraints
        """

        self._validate_scenario_configs()
        self._validate_ram()

    def _validate_scenario_configs(self) -> None:
        """Validates ScenarioConfigs

        prevents duplicate scenario labels and ensures no mixups using BGPFull
        """

        scenario_labels = list()
        for scenario_config in self.scenario_configs:
            scenario_labels.append(scenario_config.scenario_label)
            if isinstance(scenario_config.AdoptPolicyCls, BGPFull) and not isinstance(
                scenario_config.BasePolicyCls, BGPFull
            ):
                raise TypeError(
                    "You have an AdoptPolicyCls inheriting from BGPFull "
                    "but your BasePolicyCls is not. You may want to pass in "
                    "BasePolicyCls=BGPFull to your scenario_config"
                )

        if len(set(scenario_labels)) != len(scenario_labels):
            raise ValueError(
                "Each ScenarioConfig uses a scenario_label when aggregating results "
                "which by default is set to the AdoptPolicyCls's name attribute. "
                "Since you have two ScenarioConfig's with the same label, metrics "
                "won't aggregate properly. Please pass in a scenario_label= with a "
                "unique label name to your config"
            )

    def _validate_ram(self) -> None:
        """Validates that the RAM will not run out of bounds

        NOTE: all these values where obtained using pypy3.10 on a laptop
        """

        graph_kwargs = self.as_graph_constructor_kwargs.get("as_graph_kwargs", {})
        store_customer_cone_asns = graph_kwargs.get("store_customer_cone_asns", False)
        store_provider_cone_asns = graph_kwargs.get("store_provider_cone_asns", False)

        # NOTE: These are for PyPy, not Python
        # How much RAM for storing both provider and customer cones
        if store_customer_cone_asns and store_provider_cone_asns:
            total_gb_ram_per_core = 2.3
        # How much RAM for storing either provider or customer cone
        elif store_customer_cone_asns and store_provider_cone_asns:
            total_gb_ram_per_core = 1.6
        # By default sims take ~1gb/core
        else:
            total_gb_ram_per_core = 0.9

        expected_total_gb_ram = self.parse_cpus * total_gb_ram_per_core
        # Gets available RAM and converts to GB
        total_gb_ram = psutil.virtual_memory().available / (1024**3)

        bgpy_logger.info(f"Expected RAM usage: {expected_total_gb_ram:.2f}")
        bgpy_logger.info(f"Available RAM: {total_gb_ram:.2f}")
        if expected_total_gb_ram * 1.1 > total_gb_ram:
            warn(
                f"Estimated RAM usage is {expected_total_gb_ram:.2f}GB "
                f"but your machine has only {total_gb_ram:.2f}GB available, "
                "maybe use less cores or don't store provider/customer cones?",
                stacklevel=2,
            )

    def run(
        self,
        GraphFactoryCls: type[GraphFactory] | None = GraphFactory,
        graph_factory_kwargs=None,
    ) -> None:
        """Runs the simulation and write the data"""

        # Cache the CAIDA graph
        self.ASGraphConstructorCls(**self.as_graph_constructor_kwargs).run()
        graph_data_aggregator = self._get_data()
        graph_data_aggregator.write_data(
            csv_path=self.csv_path, pickle_path=self.pickle_path
        )
        self._graph_data(GraphFactoryCls, graph_factory_kwargs)
        # This object holds a lot of memory, good to get rid of it
        del graph_data_aggregator
        gc.collect()
        shutil.rmtree(self._tqdm_tracking_dir)

    def _seed_random(self, seed_suffix: str = "") -> None:
        """Seeds randomness"""

        if self.python_hash_seed is not None:
            msg = (
                f"You've set the python_hash_seed to {self.python_hash_seed}, but "
                "the simulations aren't deterministic unless you also set the "
                "PYTHONHASHSEED in the env, such as with \n"
                f"export PYTHONHASHSEED={self.python_hash_seed}"
            )
            if os.environ.get("PYTHONHASHSEED") != str(self.python_hash_seed):
                raise RuntimeError(msg)
            random.seed(str(self.python_hash_seed) + seed_suffix)

    def _get_data(self):
        """Runs trials for graph and aggregates data"""

        # Single process
        if self.parse_cpus == 1:
            # Results are a list of lists of metric trackers that we then sum
            return sum(
                self._get_single_process_results(), start=self.GraphDataAggregatorCls()
            )
        # Multiprocess
        else:
            # Results are a list of lists of metric trackers that we then sum
            return sum(self._get_mp_results(), start=self.GraphDataAggregatorCls())

    ###########################
    # Multiprocessing Methods #
    ###########################

    def _get_chunks(self, cpus: int) -> list[list[int]]:
        """Returns chunks of trial inputs based on number of CPUs running

        Not a generator since we need this for multiprocessing

        We also don't multiprocess one by one because the start up cost of
        each process is huge (since each process must generate it's own engine
        ) so we must divy up the work beforehand
        """

        trials_list = list(range(self.num_trials))
        return [trials_list[i::cpus] for i in range(cpus)]

    def _get_single_process_results(self) -> list[GraphDataAggregator]:
        """Get all results when using single processing"""

        return [self._run_chunk(i, x) for i, x in enumerate(self._get_chunks(1))]

    def _get_mp_results(self) -> list[GraphDataAggregator]:
        """Get results from multiprocessing

        Previously used starmap, but now we have tqdm
        """

        # Pool is much faster than ProcessPoolExecutor
        with Pool(self.parse_cpus) as p:
            # return p.starmap(self._run_chunk, enumerate(self._get_chunks(parse_cpus)))
            chunks = self._get_chunks(self.parse_cpus)
            desc = f"Simulating {self.output_dir.name}"
            total = sum(len(x) for x in chunks) * len(self.percent_adoptions)
            with tqdm(total=total, desc=desc) as pbar:
                tasks: list[ApplyResult[GraphDataAggregator]] = [
                    p.apply_async(self._run_chunk, x) for x in enumerate(chunks)
                ]
                completed: list[GraphDataAggregator] = []
                while tasks:
                    completed, tasks = self._get_completed_and_tasks(completed, tasks)
                    self._update_tqdm_progress_bar(pbar)
                    time.sleep(0.5)
        return completed

    def _get_completed_and_tasks(self, completed, tasks):
        """Moves completed tasks into completed"""
        new_tasks = list()
        for task in tasks:
            if task.ready():
                completed.append(task.get())
            else:
                new_tasks.append(task)
        return completed, new_tasks

    # idk what the tqdm types are supposed to be here
    def _update_tqdm_progress_bar(self, pbar: tqdm) -> None:  # type: ignore
        """Updates tqdm progress bar"""

        total_completed = 0
        for file_path in self._tqdm_tracking_dir.iterdir():
            try:
                total_completed += int(file_path.read_text())
            # Can happen write when file is being written when it's empty
            except ValueError:
                pass
        pbar.n = total_completed
        pbar.refresh()

    ############################
    # Data Aggregation Methods #
    ############################

    def _run_chunk(self, chunk_id: int, trials: list[int]) -> GraphDataAggregator:
        """Runs a chunk of trial inputs"""

        # Must also seed randomness here since we don't want multiproc to be the same
        self._seed_random(seed_suffix=str(chunk_id))

        engine = self._get_engine_for_run_chunk()

        graph_data_aggregator = self.GraphDataAggregatorCls(
            graph_categories=self.graph_categories
        )

        for trial_index, trial in self._get_run_chunk_iter(trials):
            # Use the same attacker victim pairs across all percent adoptions
            trial_attacker_asns = None
            trial_victim_asns = None
            for percent_adopt_index, percent_adopt in enumerate(self.percent_adoptions):
                # Use the same adopting asns across all scenarios configs
                adopting_asns = None
                for scenario_config in self.scenario_configs:
                    # Create the scenario for this trial
                    assert scenario_config.ScenarioCls, "ScenarioCls is None"
                    scenario = scenario_config.ScenarioCls(
                        scenario_config=scenario_config,
                        percent_adoption=percent_adopt,
                        engine=engine,
                        attacker_asns=trial_attacker_asns,
                        victim_asns=trial_victim_asns,
                        adopting_asns=adopting_asns,
                    )

                    # Change AS Classes, seed announcements before propagation
                    scenario.setup_engine(engine)
                    # For each round of propagation run the engine
                    for propagation_round in range(scenario_config.propagation_rounds):
                        self._single_engine_run(
                            engine=engine,
                            percent_adopt=percent_adopt,
                            trial=trial,
                            scenario=scenario,
                            propagation_round=propagation_round,
                            graph_data_aggregator=graph_data_aggregator,
                        )
                    if self.reuse_attacker_asns:
                        trial_attacker_asns = scenario.attacker_asns
                    if self.reuse_victim_asns:
                        trial_victim_asns = scenario.victim_asns
                    if self.reuse_adopting_asns:
                        adopting_asns = scenario.adopting_asns
                # Used to track progress with tqdm
                total_completed = (
                    trial_index * len(self.percent_adoptions) + percent_adopt_index + 1
                )
                self._write_tqdm_progress(chunk_id, total_completed)

        self._write_tqdm_progress(chunk_id, len(trials) * len(self.percent_adoptions))

        return graph_data_aggregator

    @cached_property
    def reuse_attacker_asns(self) -> bool:
        num_attackers_set = {x.num_attackers for x in self.scenario_configs}
        attacker_subcategories_set = {
            x.attacker_subcategory_attr for x in self.scenario_configs
        }
        return len(num_attackers_set) == 1 and len(attacker_subcategories_set) == 1

    @cached_property
    def reuse_victim_asns(self) -> bool:
        num_victims_set = {x.num_victims for x in self.scenario_configs}
        victim_subcategories_set = {
            x.victim_subcategory_attr for x in self.scenario_configs
        }
        return len(num_victims_set) == 1 and len(victim_subcategories_set) == 1

    @cached_property
    def reuse_adopting_asns(self) -> bool:
        adoption_categories_set = {
            x.adoption_subcategory_attrs for x in self.scenario_configs
        }
        return len(adoption_categories_set) == 1

    def _get_engine_for_run_chunk(self) -> BaseSimulationEngine:
        """Returns SimulationEngine for the _run_chunk method

        engine isn't picklable or dillable, as it has weakrefs, which
        will deserialize to dead refs
        """
        constructor_kwargs = dict(self.as_graph_constructor_kwargs)
        constructor_kwargs["tsv_path"] = None
        as_graph: ASGraph = self.ASGraphConstructorCls(**constructor_kwargs).run()
        engine = self.SimulationEngineCls(
            as_graph,
            cached_as_graph_tsv_path=self.as_graph_constructor_kwargs.get("tsv_path"),
        )
        return engine

    def _get_run_chunk_iter(self, trials: list[int]) -> Iterable[tuple[int, int]]:
        """Returns iterator for trials with or without progress bar

        If there's only 1 cpu, run the progress bar here, else we run it elsewhere
        """

        if self.parse_cpus == 1:
            return tqdm(
                enumerate(trials),
                total=len(trials),
                desc=f"Simulating {self.output_dir.name}",
            )
        else:
            return enumerate(trials)

    def _write_tqdm_progress(self, chunk_id: int, completed: int) -> None:
        """Writes total number of percent adoption trial pairs to file"""

        # If self.parse_cpus == 1, then no multiprocessing is used
        if self.parse_cpus > 1:
            with (self._tqdm_tracking_dir / f"{chunk_id}.txt").open("w") as f:
                f.write(str(completed))

    def _single_engine_run(
        self,
        *,
        engine: BaseSimulationEngine,
        percent_adopt: float | SpecialPercentAdoptions,
        trial: int,
        scenario: Scenario,
        propagation_round: int,
        graph_data_aggregator: GraphDataAggregator,
    ) -> None:
        """Single engine run"""

        # Run the engine
        engine.run(propagation_round=propagation_round, scenario=scenario)

        # Pre-aggregation Hook
        scenario.pre_aggregation_hook(
            engine=engine,
            percent_adopt=percent_adopt,
            trial=trial,
            propagation_round=propagation_round,
        )

        self._collect_engine_run_data(
            engine,
            percent_adopt,
            trial,
            scenario,
            propagation_round,
            graph_data_aggregator,
        )

        # By default, this is a no op
        scenario.post_propagation_hook(
            engine=engine,
            percent_adopt=percent_adopt,
            trial=trial,
            propagation_round=propagation_round,
        )

    def _collect_engine_run_data(
        self,
        engine: SimulationEngine,
        percent_adopt: float | SpecialPercentAdoptions,
        trial: int,
        scenario: Scenario,
        propagation_round: int,
        graph_data_aggregator: GraphDataAggregator,
    ) -> dict[int, dict[int, int]]:
        # Save all engine run info
        # The reason we aggregate info right now, instead of saving
        # the engine and doing it later, is because doing it all
        # in RAM is MUCH faster, and speed is important
        outcomes = self.ASGraphAnalyzerCls(
            engine=engine,
            scenario=scenario,
            data_plane_tracking=self.data_plane_tracking,
            control_plane_tracking=self.control_plane_tracking,
        ).analyze()

        graph_data_aggregator.aggregate_and_store_trial_data(
            engine=engine,
            percent_adopt=percent_adopt,
            trial=trial,
            scenario=scenario,
            propagation_round=propagation_round,
            outcomes=outcomes,
        )
        return outcomes

    ######################
    # Data Writing Funcs #
    ######################

    @property
    def csv_path(self) -> Path:
        return self.output_dir / "data.csv"

    @property
    def pickle_path(self) -> Path:
        return self.output_dir / "data.pickle"

    #######################
    # Graph Writing Funcs #
    #######################

    def _graph_data(
        self,
        GraphFactoryCls: type[GraphFactory] | None = GraphFactory,
        kwargs=None,
    ) -> None:
        """Generates some default graphs"""

        # This prevents problems if kwargs is reused more than once
        # outside of this file, since we modify it
        kwargs = dict() if kwargs is None else deepcopy(kwargs)
        # Set defaults for kwargs
        kwargs["pickle_path"] = kwargs.pop("pickle_path", self.pickle_path)
        kwargs["graph_dir"] = kwargs.pop("graph_dir", self.output_dir / "graphs")
        if GraphFactoryCls:
            GraphFactoryCls(**kwargs).generate_graphs()
            bgpy_logger.info(f"\nWrote graphs to {kwargs['graph_dir']}")

    @property
    def graph_output_dir(self) -> Path:
        return self.output_dir / "graphs"



from enum import Enum, unique

yamlable_enums: list[type["YamlAbleEnum"]] = []


# Yaml must have unique keys/values
@unique
class YamlAbleEnum(Enum):
    def __init_subclass__(cls: type["YamlAbleEnum"], *args, **kwargs) -> None:
        """This method essentially creates a list of all subclasses

        This is used later in the yaml codec
        """

        super().__init_subclass__(*args, **kwargs)
        yamlable_enums.append(cls)

    @classmethod
    def yaml_suffix(cls: type["YamlAbleEnum"]) -> str:
        return cls.__name__

    @staticmethod
    def yamlable_enums() -> list[type["YamlAbleEnum"]]:
        return yamlable_enums


class Outcomes(YamlAbleEnum):
    ATTACKER_SUCCESS: int = 0
    VICTIM_SUCCESS: int = 1
    DISCONNECTED: int = 2
    UNDETERMINED: int = 3


class Relationships(YamlAbleEnum):
    # Must start at one for the priority
    PROVIDERS: int = 1
    PEERS: int = 2
    # Customers have highest priority
    # Economic incentives first!
    CUSTOMERS: int = 3
    # Origin must always remain
    ORIGIN: int = 4
    # Unknown for external programs like extrapoaltor
    UNKNOWN: int = 5


class Plane(YamlAbleEnum):
    # Changing to integers so that this is compatible with c++
    DATA: int = 0  # "data_plane"
    CTRL: int = 1  # "control_plane"


class ROAValidity(YamlAbleEnum):
    """Possible values for ROA Validity

    Note that we cannot differentiate between
    invalid by origin or max length
    because you could get one that is invalid by origin for one roa
    and invalid by max length for another roa
    """

    VALID: int = 0
    UNKNOWN: int = 1
    INVALID: int = 2


class Timestamps(YamlAbleEnum):
    """Different timestamps to use"""

    # Victim is always first
    VICTIM: int = 0
    ATTACKER: int = 1


class Prefixes(YamlAbleEnum):
    """Prefixes to use for attacks

    prefix always belongs to the victim
    """

    SUPERPREFIX: str = "1.0.0.0/8"
    # Prefix always belongs to victim
    PREFIX: str = "1.2.0.0/16"
    SUBPREFIX: str = "1.2.3.0/24"


class ASNs(YamlAbleEnum):
    """Default ASNs for various ASNs"""

    ATTACKER: int = 666
    VICTIM: int = 777


class ASGroups(YamlAbleEnum):
    """AS types"""

    IXPS: str = "ixp"
    # NOTE: only the IXP group has IXPs
    STUBS: str = "stub"
    MULTIHOMED: str = "multihomed"
    STUBS_OR_MH: str = "stub_or_multihomed"
    INPUT_CLIQUE: str = "input_clique"
    # Not stubs, multihomed, or input clique
    ETC: str = "etc"
    # not stubs or multihomed
    TRANSIT: str = "transit"
    ALL_WOUT_IXPS: str = "all_wout_ixps"


class SpecialPercentAdoptions(YamlAbleEnum):
    ALL_BUT_ONE: float = 1
    ONLY_ONE: float = 0

    def __float__(self) -> float:
        return float(self.value)

    def __lt__(self, other):
        if isinstance(other, (SpecialPercentAdoptions, float)):
            return float(self) == float(other)
        else:
            return NotImplemented


class InAdoptingASNs(YamlAbleEnum):
    TRUE: str = "True"
    FALSE: str = "False"
    ANY: str = "Any"



import csv
import pickle
from collections import defaultdict
from math import sqrt
from pathlib import Path
from statistics import mean, stdev
from typing import Any
from warnings import warn

from bgpy.shared.enums import Outcomes, Plane, SpecialPercentAdoptions
from bgpy.simulation_engine import BaseSimulationEngine
from bgpy.simulation_framework.scenarios import Scenario
from bgpy.simulation_framework.utils import get_all_graph_categories

from .data_point_agg_data import DataPointAggData
from .data_point_key import DataPointKey
from .graph_category import GraphCategory
from .trial_data import TrialData

DATA_TYPE = dict[GraphCategory, defaultdict[DataPointKey, list[float]]]
PICKLE_DATA_TYPE = dict[GraphCategory, dict[DataPointKey, DataPointAggData]]


class GraphDataAggregator:
    """Aggregates data for all the graphs"""

    def __init__(
        self,
        data: DATA_TYPE | None = None,
        graph_categories: tuple[GraphCategory, ...] = tuple(get_all_graph_categories()),
    ) -> None:
        """Inits data"""

        # Data is the key for single data point on a graph
        # key DataKey (prop_round, percent_adopt, scenario_label, MetricKey)
        # metric_key contains filtering info for the type of graph/data we collect
        if data:
            self.data: DATA_TYPE = data
        else:
            self.data = {x: defaultdict(list) for x in graph_categories}

        self.graph_categories: tuple[GraphCategory, ...] = graph_categories

    #############
    # Add Funcs #
    #############

    def __add__(self, other: Any) -> "GraphDataAggregator":
        """Merges other GraphDataAggregator into this one and combines the data

        This gets called when we need to merge all the GraphDataAggregators
        from the various processes that were spawned
        """

        if isinstance(other, GraphDataAggregator):
            err = "All processes should use the same graph categories?"
            assert other.graph_categories == self.graph_categories, err
            # Time trials show that building a new dict here makes the most sense
            new_data: DATA_TYPE = {x: defaultdict(list) for x in self.graph_categories}
            for obj in (self, other):
                for graph_category, data_dict in obj.data.items():
                    for data_point_key, percents in data_dict.items():
                        new_data[graph_category][data_point_key].extend(percents)
            return self.__class__(data=new_data)
        else:
            return NotImplemented

    def __radd__(self, other):
        return self.__add__(other)

    ######################
    # Track Metric Funcs #
    ######################

    def aggregate_and_store_trial_data(
        self,
        *,
        engine: BaseSimulationEngine,
        percent_adopt: float | SpecialPercentAdoptions,
        trial: int,
        scenario: Scenario,
        propagation_round: int,
        outcomes: dict[int, dict[int, int]],
    ) -> None:
        """see desc

        Remember, each graph category represents a graph
        (Ex: Data plane, all ASes, attacker success, not adopting)
        Each DataPointKey represents a single point on that graph
        (Ex: 50% adoption of ROV against a prefix hijack)
        Each TrialData represents a single trial for a graph category
        and data point key comobo (gathered from all ases for that trial)

        The reason we don't simply save the engine to track metrics later
        is because the engines are very large and this would take a lot longer
        """

        # Each tracks trial data for a single graph category
        trial_datas = [TrialData(x) for x in self.graph_categories]
        self._aggregate_trial_data(
            trial_datas=trial_datas, engine=engine, scenario=scenario, outcomes=outcomes
        )

        data_point_key = DataPointKey(
            propagation_round=propagation_round,
            percent_adopt=percent_adopt,
            scenario_config=scenario.scenario_config,
        )

        for trial_data in trial_datas:
            # Cpmvert trial data as a percent
            # Ex: # of ASes attacker success and adopting / adopting
            percent = trial_data.get_percent()
            # If there are no ASes tracked in this graph category,
            # percent is None, and there's nothing to track
            if percent is not None:
                self.data[trial_data.graph_category][data_point_key].append(percent)

    def _aggregate_trial_data(
        self,
        *,
        trial_datas: list[TrialData],
        engine: BaseSimulationEngine,
        scenario: Scenario,
        outcomes: dict[int, dict[int, int]],
    ) -> None:
        """Gets data from every AS and stores it in the trial datas"""

        ctrl_plane_outcomes = outcomes[Plane.CTRL.value]
        data_plane_outcomes = outcomes[Plane.DATA.value]

        # Don't count these!
        uncountable_asns = scenario.untracked_asns

        for as_obj in engine.as_graph:
            # Don't count preset ASNs
            if as_obj.asn in uncountable_asns:
                continue
            for trial_data in trial_datas:
                # Must use .get, since if this tracking is turned off,
                # this will be an empty dict
                ctrl_plane_outcome = ctrl_plane_outcomes.get(
                    as_obj.asn, Outcomes.UNDETERMINED.value
                )
                data_plane_outcome = data_plane_outcomes.get(
                    as_obj.asn, Outcomes.UNDETERMINED.value
                )

                trial_data.add_data(
                    as_obj=as_obj,
                    engine=engine,
                    scenario=scenario,
                    ctrl_plane_outcome=ctrl_plane_outcome,
                    data_plane_outcome=data_plane_outcome,
                )

    ######################
    # Data Writing Funcs #
    ######################

    def write_data(
        self,
        csv_path: Path,
        pickle_path: Path,
    ) -> None:
        """Writes data to CSV and pickles it"""

        with csv_path.open("w") as f:
            rows = self.get_csv_rows()
            writer = csv.DictWriter(f, fieldnames=list(rows[0].keys()))
            writer.writeheader()
            writer.writerows(rows)

        with pickle_path.open("wb") as f:
            pickle.dump(self.get_pickle_data(), f)

    def get_csv_rows(self) -> list[dict[str, Any]]:
        """Returns rows for a CSV"""

        rows = list()
        for graph_category, data_dict in self.data.items():
            for data_point_key, percent_list in data_dict.items():
                # If there are no percents listed, then nothing was tracked for this
                # data point and we shouldn't store it at all, lest it mess us up later
                if self._data_is_storable(percent_list, data_point_key, graph_category):
                    row = {
                        "scenario_cls": (
                            data_point_key.scenario_config.ScenarioCls.__name__
                        ),
                        "AdoptingPolicyCls": (
                            data_point_key.scenario_config.AdoptPolicyCls.__name__
                        ),
                        "BasePolicyCls": (
                            data_point_key.scenario_config.BasePolicyCls.__name__
                        ),
                        "in_adopting_asns": graph_category.in_adopting_asns.value,
                        "outcome_type": graph_category.plane.name,
                        "as_group": graph_category.as_group.value,
                        "outcome": graph_category.outcome.name,
                        "percent_adopt": data_point_key.percent_adopt,
                        "propagation_round": data_point_key.propagation_round,
                        # percent_list can sometimes be empty
                        # for example, if we have 1 adopting AS for stubs_and_multihomed
                        # and that AS is multihomed, and not a stub, then for stubs,
                        # no ASes adopt, and trial_data is empty
                        # This is the proper way to do it,
                        # rather than defaulting trial_data
                        # to [0], which skews results when aggregating trials
                        "value": mean(percent_list) if percent_list else None,
                        "yerr": self._get_yerr(percent_list),
                        "scenario_config_label": (
                            data_point_key.scenario_config.csv_label
                        ),
                        "scenario_label": data_point_key.scenario_config.scenario_label,
                    }
                    rows.append(row)
        return rows

    def get_pickle_data(self) -> PICKLE_DATA_TYPE:
        agg_data: PICKLE_DATA_TYPE = {x: dict() for x in self.graph_categories}
        for graph_category, data_dict in self.data.items():
            for data_point_key, percent_list in data_dict.items():
                # If there are no percents listed, then nothing was tracked for this
                # data point and we shouldn't store it at all, lest it mess us up later
                if self._data_is_storable(percent_list, data_point_key, graph_category):
                    agg_data[graph_category][data_point_key] = DataPointAggData(
                        value=mean(percent_list),
                        yerr=self._get_yerr(percent_list),
                        data_point_key=data_point_key,
                    )
        return agg_data

    def _data_is_storable(
        self,
        percent_list: list[float],
        data_point_key: DataPointKey,
        graph_category: GraphCategory,
    ) -> bool:
        """Determines if there is data to be stored, else raise warning

        If there are no percents listed, then nothing was tracked for this
        data point and we shouldn't store it at all, lest it mess us up later
        """

        if percent_list:
            return True
        else:
            warn(
                f"GraphCategory {graph_category} "
                f"had no data for DataPointKey {data_point_key}. "
                "This means on that graph, for that data point, zero ASes "
                " fell into that category, regardless of outcome, "
                "so nothing was tracked or plotted for that data point. "
                "It's possible that this is fine, although it is extremely unusual",
                stacklevel=2,
            )
            return False

    def _get_yerr(self, percent_list: list[float]) -> float:
        """Returns 90% confidence interval for graphing"""

        if len(percent_list) > 1:
            yerr_num = 1.645 * 2 * stdev(percent_list)
            yerr_denom = sqrt(len(percent_list))
            return float(yerr_num / yerr_denom)
        else:
            return 0



from .diagram import Diagram
from .engine_run_config import EngineRunConfig
from .engine_runner import EngineRunner
from .simulator_codec import SimulatorCodec

__all__ = [
    "Diagram",
    "EngineRunConfig",
    "EngineRunner",
    "SimulatorCodec",
]



from typing import TYPE_CHECKING

from bgpy.simulation_engine.policies.rov import ROV

if TYPE_CHECKING:
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine import Announcement as Ann


class PathEnd(ROV):
    """An Policy that deploys Path-End"""

    name: str = "Path-End"

    def _valid_ann(self, ann: "Ann", recv_rel: "Relationships") -> bool:
        """Returns announcement validity by checking pathend records"""

        origin_asn = ann.origin
        origin_as_obj = self.as_.as_graph.as_dict[origin_asn]
        # If the origin is deploying pathend and the path is longer than 1
        if isinstance(origin_as_obj.policy, PathEnd) and len(ann.as_path) > 1:
            # If the provider is real, do the loop check
            # Mypy thinks this is unreachable for some reason, even tho tests pass
            for neighbor in origin_as_obj.neighbors:
                if neighbor.asn == ann.as_path[-2]:
                    return super()._valid_ann(ann, recv_rel)
            # Provider is fake, return False
            return False
        else:
            return super()._valid_ann(ann, recv_rel)



from typing import TYPE_CHECKING, Any, Optional

from bgpy.shared.enums import Relationships

from .base_simulation_engine import BaseSimulationEngine

# https://stackoverflow.com/a/57005931/8903959
if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_framework import Scenario


class SimulationEngine(BaseSimulationEngine):
    """Python simulation engine representation"""

    ###############
    # Setup funcs #
    ###############

    def setup(self, scenario: "Scenario") -> None:
        """Sets AS classes and seeds announcements"""

        self._set_as_classes(scenario)
        self._seed_announcements(scenario.announcements)
        self.ready_to_run_round = 0

    def _set_as_classes(self, scenario: "Scenario") -> None:
        """Resets Engine ASes and changes their AS class

        We do this here because we already seed from the scenario
        to allow for easy overriding. If scenario controls seeding,
        it doesn't make sense for engine to control resetting either
        and have each do half and half
        """

        # Done here to save as much time  as possible
        for as_obj in self.as_graph:
            # Delete the old policy and remove references so that RAM can be reclaimed
            del as_obj.policy.as_
            # set the AS class to be the proper type of AS
            Cls = scenario.get_policy_cls(as_obj)
            as_obj.policy = Cls(as_=as_obj)

    def _seed_announcements(self, announcements: tuple["Ann", ...] = ()) -> None:
        """Seeds announcement at the proper AS

        Since this is the simulator engine, we should
        never have to worry about overlapping announcements
        """

        for ann in announcements:
            assert ann.seed_asn is not None
            # Get the AS object to seed at
            obj_to_seed = self.as_graph.as_dict[ann.seed_asn]
            obj_to_seed.policy.seed_ann(ann)

    #####################
    # Propagation funcs #
    #####################

    def run(self, propagation_round: int = 0, scenario: Optional["Scenario"] = None):
        """Propogates announcements and ensures proper setup"""

        # Ensure that the simulator is ready to run this round
        if self.ready_to_run_round != propagation_round:
            raise RuntimeError(
                f"Engine not set up to run for {propagation_round} round"
            )
        assert scenario, "This can't be empty"

        # Propogate anns
        self._propagate(propagation_round, scenario)
        # Increment the ready to run round
        self.ready_to_run_round += 1

    def _propagate(self, propagation_round: int, scenario: "Scenario"):
        """Propogates announcements

        to stick with Gao Rexford, we propagate to
        0. providers
        2. peers
        3. customers
        """

        self._propagate_to_providers(propagation_round, scenario)
        self._propagate_to_peers(propagation_round, scenario)
        self._propagate_to_customers(propagation_round, scenario)

    def _propagate_to_providers(self, propagation_round: int, scenario: "Scenario"):
        """Propogate to providers"""

        # Propogation ranks go from stubs to input_clique in ascending order
        # By customer provider pairs (peers are ignored for the ranks)
        for i, rank in enumerate(self.as_graph.propagation_ranks):
            # Nothing to process at the start
            if i > 0:
                # Process first because maybe it recv from lower ranks
                for as_obj in rank:
                    as_obj.policy.process_incoming_anns(
                        from_rel=Relationships.CUSTOMERS,
                        propagation_round=propagation_round,
                        scenario=scenario,
                    )
            # Send to the higher ranks
            for as_obj in rank:
                as_obj.policy.propagate_to_providers()

    def _propagate_to_peers(
        self, propagation_round: int, scenario: Optional["Scenario"]
    ):
        """Propagate to peers"""

        # The reason you must separate this for loop here
        # is because propagation ranks do not take into account peering
        # It'd be impossible to take into account peering
        # since different customers peer to different ranks
        # So first do customer to provider propagation, then peer propagation
        for as_obj in self.as_graph:
            as_obj.policy.propagate_to_peers()
        for as_obj in self.as_graph:
            as_obj.policy.process_incoming_anns(
                from_rel=Relationships.PEERS,
                propagation_round=propagation_round,
                scenario=scenario,
            )

    def _propagate_to_customers(self, propagation_round: int, scenario: "Scenario"):
        """Propagate to customers"""

        # Propogation ranks go from stubs to input_clique in ascending order
        # By customer provider pairs (peers are ignored for the ranks)
        # So here we start at the highest rank(input_clique) and propagate down
        for i, rank in enumerate(reversed(self.as_graph.propagation_ranks)):
            # There are no incomming Anns at the top
            if i > 0:
                for as_obj in rank:
                    as_obj.policy.process_incoming_anns(
                        from_rel=Relationships.PROVIDERS,
                        propagation_round=propagation_round,
                        scenario=scenario,
                    )
            for as_obj in rank:
                as_obj.policy.propagate_to_customers()

    ##############
    # Yaml funcs #
    ##############

    def __to_yaml_dict__(self) -> dict[str, Any]:
        """This optional method is called when you call yaml.dump()"""

        return dict(vars(self))

    @classmethod
    def __from_yaml_dict__(
        cls: type["SimulationEngine"], dct: dict[str, Any], yaml_tag: Any
    ) -> "SimulationEngine":
        """This optional method is called when you call yaml.load()"""

        return cls(**dct)



from ipaddress import IPv4Network, IPv6Network, ip_network
from typing import TYPE_CHECKING

from roa_checker import ROAChecker, ROAValidity

from bgpy.simulation_engine import Announcement as Ann
from bgpy.simulation_engine import BaseSimulationEngine

"""ROA helper funcs. I would move more but that would break mypy"""

if TYPE_CHECKING:
    from .scenario import Scenario


def _add_roa_info_to_anns(
    self: "Scenario",
    *,
    announcements: tuple["Ann", ...] = (),
    engine: BaseSimulationEngine | None = None,
) -> tuple["Ann", ...]:
    """Adds ROA Info to Announcements"""

    if self.roas:
        roa_checker = self._get_roa_checker()
        processed_anns = list()
        for ann in announcements:
            prefix = ip_network(ann.prefix)

            roa_origin = self._get_roa_origin(roa_checker, prefix, ann.origin)

            roa_valid_length = self._get_roa_valid_length(
                roa_checker, prefix, ann.origin
            )

            processed_anns.append(
                ann.copy(
                    {
                        "roa_valid_length": roa_valid_length,
                        "roa_origin": roa_origin,
                    }
                )
            )
        return tuple(processed_anns)
    else:
        return announcements


def _get_roa_checker(self: "Scenario") -> ROAChecker:
    """Returns ROAChecker populated with self.roas"""

    roa_checker = ROAChecker()
    for roa in self.roas:
        roa_checker.insert(roa.prefix, roa)
    return roa_checker


def _get_roa_origin(
    self: "Scenario",
    roa_checker: ROAChecker,
    prefix: IPv4Network | IPv6Network,
    origin: int,
) -> int | None:
    """Returns ROA origin"""

    roas = roa_checker.get_relevant_roas(prefix)
    if len(roas) == 0:
        return None
    elif len(roas) == 1:
        [roa] = roas
        return int(roa.origin)
    else:
        raise NotImplementedError


def _get_roa_valid_length(
    self: "Scenario",
    roa_checker: ROAChecker,
    prefix: IPv4Network | IPv6Network,
    origin: int,
) -> bool | None:
    """Returns ROA validity"""

    outcome = roa_checker.get_roa_outcome(prefix, origin)
    if outcome.validity in (
        ROAValidity.INVALID_LENGTH,
        ROAValidity.INVALID_LENGTH_AND_ORIGIN,
    ):
        return False
    elif outcome.validity == ROAValidity.UNKNOWN:
        return None
    elif outcome.validity in (ROAValidity.VALID, ROAValidity.INVALID_ORIGIN):
        return True
    else:
        raise NotImplementedError(f"Should never reach this {outcome.validity}")



from .first_asn_stripping_prefix_aspa_attacker import (
    FirstASNStrippingPrefixASPAAttacker,
)
from .shortest_path_prefix_aspa_attacker import ShortestPathPrefixASPAAttacker

__all__ = ["FirstASNStrippingPrefixASPAAttacker", "ShortestPathPrefixASPAAttacker"]



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import ASPA, BGP
from bgpy.simulation_framework import AccidentalRouteLeak, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = "accidental route leak against ASPASimple"

ex_config_021 = EngineTestConfig(
    name="ex_021_route_leak_aspa_simple_upstream_verification",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=AccidentalRouteLeak,
        BasePolicyCls=BGP,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                1: ASPA,
                2: ASPA,
            }
        ),
    ),
    as_graph_info=as_graph_info_000,
)



from bgpy.simulation_engine.policies.bgp.bgp_full import BGPFull

from .bgpsec import BGPSec


class BGPSecFull(BGPSec, BGPFull):
    """Represents BGPSec with withdrawals, ribsin, ribs out"""

    name: str = "BGPsec Full"



from typing import TYPE_CHECKING

from bgpy.shared.enums import Relationships
from bgpy.simulation_engine.policies.rovpp.v2.base.rovpp_v2_lite import ROVPPV2Lite

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann


class ROVPPV2ImprovedLite(ROVPPV2Lite):
    """An Policy that deploys ROV++V2 Lite as defined in the ROV++ paper

    With a deviation:
    We published that an AS that recieves a hijack from it's customers
    shouldn't also send to it's customers

    Really though, this makes no sense, since you can protect your other
    customers from this by sending a competing blackhole

    When we tested this, it was slightly better than the default v2. We
    planned on publishing this in an ROV++ Journal version but scrapped
    the paper to work on other things. Have fun!

    ROV++ Improved Deployable Defense against BGP Hijacking
    """

    name: str = "ROV++V2i Lite"

    def _send_competing_hijack_allowed(
        self, ann: "Ann", propagate_to: Relationships
    ) -> bool:
        """You can send blackhole if either subprefix or non routed

        Unlike our conference paper, we don't care where the original
        hijack came from (which performs better)
        """

        return (
            # We don't care about this part (default v2), it doesn't help
            # From peer/provider
            # ann.recv_relationship
            # in [Relationships.PEERS, Relationships.PROVIDERS, Relationships.ORIGIN]
            # Sending to customers
            propagate_to == Relationships.CUSTOMERS
            # subprefix or non routed (don't send blackholes for prefixes)
            and (not ann.roa_valid_length or not ann.roa_routed)
        )



from pathlib import Path

from bgpy.shared.enums import SpecialPercentAdoptions
from bgpy.simulation_engine import ROV
from bgpy.simulation_framework import ScenarioConfig, Simulation, SubprefixHijack


def main():
    """Runs the defaults"""

    # Simulation for the paper
    sim = Simulation(
        percent_adoptions=(
            SpecialPercentAdoptions.ONLY_ONE,
            0.1,
            0.2,
            0.5,
            0.8,
            0.99,
            # Using only 1 AS not adopting causes extreme variance
            # SpecialPercentAdoptions.ALL_BUT_ONE,
        ),
        scenario_configs=(
            ScenarioConfig(ScenarioCls=SubprefixHijack, AdoptPolicyCls=ROV),
        ),
        output_dir=Path("~/Desktop/main_ex").expanduser(),
        num_trials=10,
        parse_cpus=10,
    )
    sim.run()


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

from bgpy.shared.enums import Relationships

if TYPE_CHECKING:
    from bgpy.as_graphs import AS
    from bgpy.simulation_engine.announcement import Announcement as Ann

    from .bgp import BGP


def propagate_to_providers(self) -> None:
    """Propogates to providers

    Propogate ann's that have a recv_rel of origin or customer to providers
    """

    send_rels: set[Relationships] = {
        Relationships.ORIGIN,
        Relationships.CUSTOMERS,
    }
    self._propagate(Relationships.PROVIDERS, send_rels)


def propagate_to_customers(self) -> None:
    """Propogates to customers"""

    # Anns that have any of these as recv_rel get propogated
    send_rels: set[Relationships] = {
        Relationships.ORIGIN,
        Relationships.CUSTOMERS,
        Relationships.PEERS,
        Relationships.PROVIDERS,
    }
    self._propagate(Relationships.CUSTOMERS, send_rels)


def propagate_to_peers(self) -> None:
    """Propogates to peers"""

    # Anns that have any of these as recv_rel get propogated
    send_rels: set[Relationships] = {
        Relationships.ORIGIN,
        Relationships.CUSTOMERS,
    }
    self._propagate(Relationships.PEERS, send_rels)


def _propagate(
    self: "BGP",
    propagate_to: Relationships,
    send_rels: set[Relationships],
) -> None:
    """Propogates announcements from local rib to other ASes

    send_rels is the relationships that are acceptable to send
    """

    if propagate_to.value == Relationships.PROVIDERS.value:
        neighbors = self.as_.providers
    elif propagate_to.value == Relationships.PEERS.value:
        neighbors = self.as_.peers
    elif propagate_to.value == Relationships.CUSTOMERS.value:
        neighbors = self.as_.customers
    else:
        raise NotImplementedError

    for _prefix, unprocessed_ann in self.local_rib.items():
        # Starting in v4 we must set the next_hop when sending
        # Copying announcements is a bottleneck for sims,
        # so we try to do this as little as possible
        if neighbors and unprocessed_ann.recv_relationship in send_rels:
            ann = unprocessed_ann.copy({"next_hop_asn": self.as_.asn})
        else:
            continue

        for neighbor in neighbors:
            if ann.recv_relationship in send_rels and not self._prev_sent(
                neighbor, ann
            ):
                # Policy took care of it's own propagation for this ann
                if self._policy_propagate(neighbor, ann, propagate_to, send_rels):
                    continue
                else:
                    self._process_outgoing_ann(neighbor, ann, propagate_to, send_rels)


def _policy_propagate(
    self: "BGP",
    neighbor: "AS",
    ann: "Ann",
    propagate_to: Relationships,
    send_rels: set[Relationships],
) -> bool:
    """Custom policy propagation that can be overriden"""

    return False


def _prev_sent(self: "BGP", neighbor: "AS", ann: "Ann") -> bool:
    """Don't resend anything for BGPAS. For this class it doesn't matter"""
    return False


def _process_outgoing_ann(
    self: "BGP",
    neighbor: "AS",
    ann: "Ann",
    propagate_to: Relationships,
    send_rels: set[Relationships],
) -> None:
    """Adds ann to the neighbors recv q"""

    # Add the new ann to the incoming anns for that prefix
    neighbor.policy.receive_ann(ann)



from bgpy.simulation_engine.policies.bgp.bgp_full import BGPFull

from .only_to_customers import OnlyToCustomers


class OnlyToCustomersFull(OnlyToCustomers, BGPFull):
    """An Policy that deploys OnlyToCustomers and has withdrawals, ribs in and out"""

    name: str = "OnlyToCustomers Full"



from typing import TYPE_CHECKING, Any, Optional
from warnings import warn

from bgpy.simulation_engine.ann_containers import LocalRIB, RecvQueue
from bgpy.simulation_engine.policies.policy import Policy

# Gao rexford functions
from .gao_rexford import (
    _get_best_ann_by_as_path,
    _get_best_ann_by_gao_rexford,
    _get_best_ann_by_local_pref,
    _get_best_ann_by_lowest_neighbor_asn_tiebreaker,
)

# Process incoming announcements
from .process_incoming_funcs import (
    _copy_and_process,
    _reset_q,
    _valid_ann,
    process_incoming_anns,
    receive_ann,
    seed_ann,
)

# Propagation functionality
from .propagate_funcs import (
    _policy_propagate,
    _prev_sent,
    _process_outgoing_ann,
    _propagate,
    propagate_to_customers,
    propagate_to_peers,
    propagate_to_providers,
)

if TYPE_CHECKING:
    from weakref import CallableProxyType

    from bgpy.as_graphs import AS


class BGP(Policy):
    name: str = "BGP"

    def __init__(
        self,
        local_rib: LocalRIB | None = None,
        recv_q: RecvQueue | None = None,
        as_: Optional["AS"] = None,
    ) -> None:
        """Add local rib and data structures here

        This way they can be easily cleared later without having to redo
        the graph

        This is also useful for regenerating an AS from YAML
        """

        self.local_rib = local_rib if local_rib else LocalRIB()
        self.recv_q = recv_q if recv_q else RecvQueue()
        # This gets set within the AS class so it's fine
        self.as_: CallableProxyType[AS] = as_  # type: ignore

    # Propagation functionality
    propagate_to_providers = propagate_to_providers
    propagate_to_customers = propagate_to_customers
    propagate_to_peers = propagate_to_peers
    _propagate = _propagate
    _policy_propagate = _policy_propagate
    _process_outgoing_ann = _process_outgoing_ann
    _prev_sent = _prev_sent

    # Process incoming announcements
    seed_ann = seed_ann
    receive_ann = receive_ann
    process_incoming_anns = process_incoming_anns
    _valid_ann = _valid_ann
    _copy_and_process = _copy_and_process
    _reset_q = _reset_q

    # Gao rexford functions
    _get_best_ann_by_gao_rexford = _get_best_ann_by_gao_rexford
    _get_best_ann_by_local_pref = _get_best_ann_by_local_pref
    _get_best_ann_by_as_path = _get_best_ann_by_as_path
    _get_best_ann_by_lowest_neighbor_asn_tiebreaker = (
        _get_best_ann_by_lowest_neighbor_asn_tiebreaker
    )

    @property
    def _local_rib(self) -> LocalRIB:
        warn(
            "Please use .local_rib instead of ._local_rib. "
            "This will be removed in a later version",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.local_rib

    @property
    def _recv_q(self) -> RecvQueue:
        warn(
            "Please use .recv_q instead of ._recv_q. "
            "This will be removed in a later version",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.recv_q

    ##############
    # Yaml funcs #
    ##############

    def __to_yaml_dict__(self) -> dict[Any, Any]:
        """This optional method is called when you call yaml.dump()"""

        return {"local_rib": self.local_rib, "recv_q": self.recv_q}

    @classmethod
    def __from_yaml_dict__(cls, dct, yaml_tag) -> Policy:
        """This optional method is called when you call yaml.load()"""

        return cls(**dct)



from dataclasses import asdict, dataclass, replace
from typing import Any, Optional

from yamlable import YamlAble, yaml_info

from bgpy.shared.enums import Relationships


@yaml_info(yaml_tag="Announcement")
@dataclass(slots=True, frozen=True)
class Announcement(YamlAble):
    """BGP Announcement"""

    prefix: str
    as_path: tuple[int, ...]
    # Equivalent to the next hop in a normal BGP announcement
    next_hop_asn: int = None  # type: ignore
    seed_asn: int | None = None
    recv_relationship: Relationships = Relationships.ORIGIN

    #############################
    # Optional attributes below #
    #############################
    # If you aren't using the policies listed below,
    # You can create an announcement class without them
    # for a much faster runtime. Announcement copying is
    # the bottleneck for BGPy, smaller announcements copy
    # much faster across the AS topology

    # This currently is unused. Depending on some results from
    # our in-progress publications it may be used in the future
    # For now, we just set the timestamp of the victim to 0,
    # and timestamp of the attacker to 1
    timestamp: int = 0
    # Used for classes derived from BGPFull
    withdraw: bool = False
    # ROV, ROV++ optional attributes
    roa_valid_length: bool | None = None
    roa_origin: int | None = None
    # BGPsec optional attributes
    # BGPsec next ASN that should receive the control plane announcement
    # NOTE: this is the opposite direction of next_hop, for the data plane
    bgpsec_next_asn: int | None = None
    bgpsec_as_path: tuple[int, ...] = ()
    # RFC 9234 OTC attribute (Used in OnlyToCustomers Policy)
    only_to_customers: int | None = None
    # ROV++ attribute
    rovpp_blackhole: bool = False

    def __post_init__(self):
        """Defaults seed_asn and next_hop_asn"""

        if self.seed_asn is None and len(self.as_path) == 1:
            object.__setattr__(self, "seed_asn", self.as_path[0])
        if self.next_hop_asn is None:
            # next hop defaults to None, messing up the type
            if len(self.as_path) == 1:  # type: ignore
                object.__setattr__(self, "next_hop_asn", self.as_path[0])
            else:
                raise ValueError("Must set next_hop_asn")

    def prefix_path_attributes_eq(self, ann: Optional["Announcement"]) -> bool:
        """Checks prefix and as path equivalency"""

        if ann is None:
            return False
        elif isinstance(ann, Announcement):
            return (ann.prefix, ann.as_path) == (self.prefix, self.as_path)
        else:
            raise NotImplementedError

    def copy(
        self, overwrite_default_kwargs: dict[Any, Any] | None = None
    ) -> "Announcement":
        """Creates a new ann with proper sim attrs"""

        if overwrite_default_kwargs:
            # Mypy says it gets this wrong
            # https://github.com/microsoft/pyright/issues/1047#issue-705124399
            return replace(self, **overwrite_default_kwargs)
        else:
            return replace(self)

    def bgpsec_valid(self, asn: int) -> bool:
        """Returns True if valid by BGPSec else False"""

        return self.bgpsec_next_asn == asn and self.bgpsec_as_path == self.as_path

    @property
    def invalid_by_roa(self) -> bool:
        """Returns True if Ann is invalid by ROA

        False means ann is either valid or unknown
        """

        # Not covered by ROA, unknown
        if self.roa_origin is None:
            return False
        else:
            return self.origin != self.roa_origin or not self.roa_valid_length

    @property
    def valid_by_roa(self) -> bool:
        """Returns True if Ann is valid by ROA

        False means ann is either invalid or unknown
        """

        # Need the bool here for mypy, ugh
        return bool(self.origin == self.roa_origin and self.roa_valid_length)

    @property
    def unknown_by_roa(self) -> bool:
        """Returns True if ann is not covered by roa"""

        return not self.invalid_by_roa and not self.valid_by_roa

    @property
    def covered_by_roa(self) -> bool:
        """Returns if an announcement has a roa"""

        return not self.unknown_by_roa

    @property
    def roa_routed(self) -> bool:
        """Returns bool for if announcement is routed according to ROA"""

        return self.roa_origin != 0

    @property
    def origin(self) -> int:
        """Returns the origin of the announcement"""

        return self.as_path[-1]

    def __str__(self) -> str:
        return f"{self.prefix} {self.as_path} {self.recv_relationship}"

    ##############
    # Yaml funcs #
    ##############

    def __to_yaml_dict__(self) -> dict[str, Any]:
        """This optional method is called when you call yaml.dump()"""

        return asdict(self)

    @classmethod
    def __from_yaml_dict__(
        cls: type["Announcement"], dct: dict[str, Any], yaml_tag: Any
    ) -> "Announcement":
        """This optional method is called when you call yaml.load()"""

        return cls(**dct)



from .base import ROVPPV2Lite, ROVPPV2LiteFull
from .improved import ROVPPV2ImprovedLite, ROVPPV2ImprovedLiteFull

__all__ = [
    "ROVPPV2Lite",
    "ROVPPV2LiteFull",
    "ROVPPV2ImprovedLite",
    "ROVPPV2ImprovedLiteFull",
]



from .ex_config_000 import ex_config_000
from .ex_config_001 import ex_config_001
from .ex_config_002 import ex_config_002
from .ex_config_003 import ex_config_003
from .ex_config_004 import ex_config_004
from .ex_config_005 import ex_config_005
from .ex_config_006 import ex_config_006
from .ex_config_007 import ex_config_007
from .ex_config_008 import ex_config_008
from .ex_config_009 import ex_config_009
from .ex_config_010 import ex_config_010
from .ex_config_011 import ex_config_011
from .ex_config_012 import ex_config_012
from .ex_config_013 import ex_config_013
from .ex_config_014 import ex_config_014
from .ex_config_015 import ex_config_015
from .ex_config_016 import ex_config_016
from .ex_config_017 import ex_config_017
from .ex_config_018 import ex_config_018
from .ex_config_019 import ex_config_019
from .ex_config_020 import ex_config_020
from .ex_config_021 import ex_config_021
from .ex_config_022 import ex_config_022
from .ex_config_023 import ex_config_023
from .ex_config_024 import ex_config_024
from .ex_config_025 import ex_config_025
from .ex_config_026 import ex_config_026
from .ex_config_027 import ex_config_027
from .ex_config_028 import ex_config_028
from .ex_config_029 import ex_config_029
from .ex_config_030 import ex_config_030

example_configs = [
    ex_config_000,
    ex_config_001,
    ex_config_002,
    ex_config_003,
    ex_config_004,
    ex_config_005,
    ex_config_006,
    ex_config_007,
    ex_config_008,
    ex_config_009,
    ex_config_010,
    ex_config_011,
    ex_config_012,
    ex_config_013,
    ex_config_014,
    ex_config_015,
    ex_config_016,
    ex_config_017,
    ex_config_018,
    ex_config_019,
    ex_config_020,
    ex_config_021,
    ex_config_022,
    ex_config_023,
    ex_config_024,
    ex_config_025,
    ex_config_026,
    ex_config_027,
    ex_config_028,
    ex_config_029,
    ex_config_030,
]

__all__ = ["example_configs"]



from frozendict import frozendict

from bgpy.as_graphs import ASGraphInfo, PeerLink
from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import BGP, OnlyToCustomers
from bgpy.simulation_framework import AccidentalRouteLeak, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

r"""Graph to test OTC from a peer

777 - 666 - 1
"""

as_graph_info = ASGraphInfo(
    peer_links=frozenset(
        [
            PeerLink(ASNs.VICTIM.value, ASNs.ATTACKER.value),
            PeerLink(ASNs.ATTACKER.value, 1),
        ]
    ),
    customer_provider_links=frozenset(),
)


internal_config_003 = EngineTestConfig(
    name="internal_003",
    desc="Accidental route leak to a peer with OTC Simple",
    scenario_config=ScenarioConfig(
        ScenarioCls=AccidentalRouteLeak,
        BasePolicyCls=BGP,
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                1: OnlyToCustomers,
                ASNs.VICTIM.value: OnlyToCustomers,
            }
        ),
    ),
    as_graph_info=as_graph_info,
)



from copy import deepcopy

from frozendict import frozendict

from bgpy.as_graphs import ASGraphInfo, PeerLink
from bgpy.as_graphs import CustomerProviderLink as CPLink
from bgpy.shared.enums import Prefixes
from bgpy.simulation_engine import BGPFull
from bgpy.simulation_framework import ScenarioConfig, ValidPrefix
from bgpy.tests.engine_tests.utils import EngineTestConfig

r"""Graph to test relationship preference

      2
     /
5 - 1 - 3
     \
      4
"""

as_graph_info = ASGraphInfo(
    peer_links=frozenset([PeerLink(1, 3), PeerLink(1, 5)]),
    customer_provider_links=frozenset(
        [
            CPLink(provider_asn=2, customer_asn=1),
            CPLink(provider_asn=1, customer_asn=4),
        ]
    ),
)


class Custom02ValidPrefix(ValidPrefix):
    """Add a better announcement in round 2 to cause withdrawal"""

    min_propagation_rounds: int = 4

    # Going to just suppress mypy err here since I don't want to rewrite Cameron's func
    def post_propagation_hook(self, engine=None, propagation_round=0, *args, **kwargs):  # type: ignore
        if propagation_round == 1:  # second round
            ann = deepcopy(
                engine.as_graph.as_dict[2].policy.local_rib.get(Prefixes.PREFIX.value)
            )
            # Add a new announcement at AS 3, which will be better than the one
            # from 2 and cause a withdrawn route by 1 to 4
            # ann.seed_asn = 3
            # ann.as_path = (3,)
            object.__setattr__(ann, "seed_asn", 3)
            object.__setattr__(
                ann,
                "as_path",
                (3,),
            )
            engine.as_graph.as_dict[3].policy.local_rib.add_ann(ann)
            Custom02ValidPrefix.victim_asns = frozenset({2, 3})

        if propagation_round == 2:  # third round
            ann = deepcopy(
                engine.as_graph.as_dict[3].policy.local_rib.get(Prefixes.PREFIX.value)
            )
            object.__setattr__(ann, "withdraw", True)
            # ann.withdraw = True
            # Remove the original announcement from 3
            # The one from 2 is now the next-best
            engine.as_graph.as_dict[3].policy.local_rib.pop(Prefixes.PREFIX.value, None)
            engine.as_graph.as_dict[3].policy.ribs_out.remove_entry(
                1, Prefixes.PREFIX.value
            )
            engine.as_graph.as_dict[3].policy.send_q.add_ann(1, ann)
            Custom02ValidPrefix.victim_asns = frozenset({2})


internal_config_002 = EngineTestConfig(
    name="internal_002",
    desc="Test withdrawal mechanism choosing next best announcement",
    scenario_config=ScenarioConfig(
        ScenarioCls=Custom02ValidPrefix,
        BasePolicyCls=BGPFull,
        override_victim_asns=frozenset({2}),
        hardcoded_asn_cls_dict=frozendict(),
    ),
    as_graph_info=as_graph_info,
)



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import ASPAFull, BGPFull
from bgpy.simulation_framework import AccidentalRouteLeak, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = "accidental route leak against ASPA"

ex_config_022 = EngineTestConfig(
    name="ex_022_route_leak_aspa_upstream_verification",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=AccidentalRouteLeak,
        BasePolicyCls=BGPFull,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                1: ASPAFull,
                2: ASPAFull,
            }
        ),
    ),
    as_graph_info=as_graph_info_000,
)



from bgpy.simulation_engine.policies.rov.rov_full import ROVFull

from .aspa import ASPA


class ASPAFull(ASPA, ROVFull):
    """An Policy that deploys ASPA and has withdrawals, ribs in and out"""

    name: str = "ASPA Full"



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import BGP
from bgpy.simulation_framework import PrefixHijack, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = "Prefix hijack with BGP Simple"

ex_config_001 = EngineTestConfig(
    name="ex_001_prefix_hijack_bgp_simple",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=PrefixHijack,
        BasePolicyCls=BGP,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(),
    ),
    as_graph_info=as_graph_info_000,
)



from typing import cast

from .link import Link


class PeerLink(Link):
    """Stores the info for a peer link"""

    def __init__(self, peer1_asn: int, peer2_asn: int):
        """Saves the link info"""

        self.__peer_asns: tuple[int, int] = cast(
            tuple[int, int], tuple(sorted([int(peer1_asn), int(peer2_asn)]))
        )
        super(PeerLink, self).__init__(peer1_asn, peer2_asn)

    def __hash__(self) -> int:
        """Hashes used in sets

        NOTE: python disables hash if __eq__ is defined so you MUST explicitly redef
        """

        return hash(self.asns)

    def __eq__(self, other) -> bool:
        if isinstance(other, PeerLink):
            return self.peer_asns == other.peer_asns
        else:
            return NotImplemented

    @property
    def peer_asns(self) -> tuple[int, int]:
        """Returns peer asns. Done this way for immutability/hashing"""

        return self.__peer_asns

    @property
    def asns(self) -> tuple[int, ...]:
        """Returns asns associated with this link"""

        return tuple(sorted(self.__peer_asns))



from dataclasses import replace
from typing import TYPE_CHECKING

from frozendict import frozendict

from bgpy.simulation_framework.graphing.line_data import LineData
from bgpy.simulation_framework.graphing.line_info import LineInfo

if TYPE_CHECKING:
    from bgpy.simulation_framework import DataPointAggData, DataPointKey, GraphCategory


def _generate_graph(
    self,
    graph_category: "GraphCategory",
    data_dict: dict["DataPointKey", "DataPointAggData"],
) -> None:
    """Writes a graph to the graph dir"""

    if not data_dict:
        return

    graph_name, line_data_dict, fig, ax = self._preprocessing_steps(
        graph_category, data_dict
    )

    (
        non_aggregated_data_dict,
        max_attacker_data_dict,
    ) = self._graph_data(ax, line_data_dict)

    self._add_legends_and_save(
        fig,
        ax,
        graph_category,
        non_aggregated_data_dict,
        max_attacker_data_dict,
        graph_name,
    )


def _graph_data(self, ax, line_data_dict: dict[str, LineData]):
    # Add all lines that aren't aggregated into a strongest attacker aggregation
    self._plot_non_aggregated_lines(ax, line_data_dict)

    (
        non_aggregated_line_data_dict,
        max_attacker_data_dict,
    ) = self._plot_strongest_attacker_lines(ax, line_data_dict)
    return non_aggregated_line_data_dict, max_attacker_data_dict


def _plot_non_aggregated_lines(self, ax, line_data_dict: dict[str, LineData]) -> None:
    """Add all lines that aren't aggregated into a strongest attacker aggregation"""

    for label, line_data in line_data_dict.items():
        # Labels to aggregate for the strongest attacker shouldn't be plotted
        if label not in self.labels_to_aggregate:
            self._plot_line_data(ax, line_data)


def _plot_strongest_attacker_lines(
    self, ax, line_data_dict: dict[str, LineData]
) -> tuple[dict[str, LineData], dict[str, LineData]]:
    # Add all lines that are aggregated
    max_attacker_data_dict: dict[str, LineData] = dict()
    for label in self.labels_to_aggregate:
        line_infos = line_data_dict.pop(label, None)
        assert line_infos, "Did you pass in the wrong strongest attacker dict?"
        max_attacker_data_dict[label] = line_infos
    # At this point max_attacker_data_dict might be something like:
    # {ASPA: (ASPAForgedOriginLineInfo, ASPAShortestPathLineInfo)

    if self.labels_to_aggregate:
        strongest_agg_dict, scatter_line_data_dict = self._get_agg_data(
            max_attacker_data_dict
        )

        agg_line_datas = self._get_agg_line_data(strongest_agg_dict)

        for agg_line_data in agg_line_datas:
            line_data_dict[agg_line_data.label] = agg_line_data
            self._plot_line_data(ax, agg_line_data)

        self._plot_scatter_plots(ax, scatter_line_data_dict)

    return line_data_dict, max_attacker_data_dict


def _get_agg_data(
    self, max_attacker_data_dict: dict[str, LineData]
) -> tuple[dict[str, dict[str, list[float]]], dict[str, LineData]]:
    # Gets all lines that will be in the scatter plot
    # So basically each line beforehand, but some X values will
    # be removed if they aren't the strongest listed
    scatter_plots: dict[str, dict[str, list[float]]] = {
        label: {"xs": [], "ys": [], "yerrs": []} for label in self.labels_to_aggregate
    }

    # Get all strongest attacker lines. No data point markers, but the
    # line itself that will be plotted
    agg_xs = next(iter(max_attacker_data_dict.values())).xs
    strongest_agg_dict: dict[str, dict[str, list[float]]] = {
        label: {"agg_xs": list(agg_xs), "agg_ys": [], "agg_yerrs": [0 for _ in agg_xs]}
        for label in self.strongest_attacker_dict
    }

    # Populate the new agg line and scatter plots
    for (
        strongest_attacker_label,
        line_infos_to_agg,
    ) in self.strongest_attacker_dict.items():
        # {agg_xs, agg_ys, agg_yerrs}
        cur_data = strongest_agg_dict[strongest_attacker_label]

        for i, x in enumerate(cur_data["agg_xs"]):
            best_label = None
            max_val = None
            new_yerr = None
            for line_info in line_infos_to_agg:
                line_data = max_attacker_data_dict[line_info.label]
                # mypy doens't understand that line_infos_to_agg > len 1
                if max_val is None or line_data.ys[i] > max_val:  # type: ignore
                    best_label = line_info.label
                    max_val = line_data.ys[i]
                    new_yerr = line_data.yerrs[i]
            assert isinstance(best_label, str), "For mypy"
            assert isinstance(max_val, float), "For mypy"
            assert isinstance(new_yerr, (float, int)), f"mypy {line_infos_to_agg}"
            cur_data["agg_ys"].append(max_val)
            scatter_plots[best_label]["xs"].append(x)
            scatter_plots[best_label]["ys"].append(max_val)
            scatter_plots[best_label]["yerrs"].append(new_yerr)

    scatter_line_data_dict = self._get_scatter_line_data_dict(
        scatter_plots, max_attacker_data_dict
    )
    return strongest_agg_dict, scatter_line_data_dict


def _get_scatter_line_data_dict(
    self,
    scatter_plots: dict[str, dict[str, list[float]]],
    max_attacker_dict: dict[str, LineData],
) -> dict[str, LineData]:
    """Converts scatter plots into proper line data for plotting"""

    label_to_marker_dict = dict()
    for line_info_tup in self.strongest_attacker_dict.values():
        for line_info in line_info_tup:
            label_to_marker_dict[line_info.strongest_attacker_legend_label] = (
                line_info.marker
            )

    scatter_line_data_dict = dict()
    for label, point_dict in scatter_plots.items():
        old_line_data = max_attacker_dict[label]

        scatter_line_data_dict[label] = LineData(
            label=old_line_data.label,
            formatted_graph_rows=None,
            # This makes the line dissapear
            line_info=replace(
                old_line_data.line_info,
                marker=label_to_marker_dict[
                    old_line_data.line_info.strongest_attacker_legend_label
                ],
                ls="solid",
                color="grey",
                extra_kwargs=frozendict(
                    {
                        **dict(
                            lw=0,
                            **old_line_data.line_info.extra_kwargs,
                        ),
                        # Marker face color
                        # Since lines are colored, make color grey (jk)
                        "mfc": old_line_data.line_info.color,
                        # Marker edge color
                        "mec": old_line_data.line_info.color,
                        # Marker size
                        # "ms": 20,
                        # "markeredgewidth": 3,
                        # Old line color
                        "ecolor": old_line_data.line_info.color,
                        "zorder": 3,
                    }
                ),
            ),
            xs=tuple(point_dict["xs"]),
            ys=tuple(point_dict["ys"]),
            yerrs=tuple(point_dict["yerrs"]),
        )
    return scatter_line_data_dict


def _get_agg_line_data(
    self, strongest_agg_dict: dict[str, dict[str, list[float]]]
) -> tuple[LineData, ...]:
    line_datas = list()
    for agg_label, agg_data_dict in strongest_agg_dict.items():
        line_datas.append(
            LineData(
                agg_label,
                formatted_graph_rows=None,
                line_info=LineInfo(
                    agg_label,
                    marker=".",
                    # Reuse the same line style and color as the aggregated lines
                    ls=self.strongest_attacker_dict[agg_label][0].ls,
                    color=self.strongest_attacker_dict[agg_label][0].color,
                    # Ms stands for marker size
                    extra_kwargs=frozendict({"zorder": 0, "ms": 0, "elinewidth": 0}),
                ),
                xs=tuple(agg_data_dict["agg_xs"]),
                ys=tuple(agg_data_dict["agg_ys"]),
                yerrs=tuple(agg_data_dict["agg_yerrs"]),
            )
        )
    return tuple(line_datas)


def _plot_line_data(self, ax, line_data: LineData):
    """Plots line data"""

    return ax.errorbar(
        line_data.xs,
        line_data.ys,
        line_data.yerrs,
        label=line_data.line_info.label,
        marker=line_data.line_info.marker,
        ls=line_data.line_info.ls,
        color=line_data.line_info.color,
        **line_data.line_info.extra_kwargs,
    )


def _plot_scatter_plots(
    self, ax, scatter_plot_line_data_dict: dict[str, LineData]
) -> None:
    """Plots scatter plots"""

    for line_data in scatter_plot_line_data_dict.values():
        self._plot_line_data(ax, line_data)



from typing import TYPE_CHECKING

from bgpy.shared.exceptions import GaoRexfordError
from bgpy.simulation_engine.announcement import Announcement as Ann

if TYPE_CHECKING:
    from .bgp import BGP


def _get_best_ann_by_gao_rexford(
    self: "BGP",
    current_ann: Ann | None,
    new_ann: Ann,
) -> Ann:
    """Determines if the new ann > current ann by Gao Rexford"""

    assert new_ann is not None, "New announcement can't be None"

    if current_ann is None:
        return new_ann
    else:
        # Inspiration for this func refactor came from bgpsecsim
        # for func in self._gao_rexford_funcs:
        #     best_ann = func(current_ann, new_ann)
        #     if best_ann is not None:
        #         assert isinstance(best_ann, Ann), "mypy type check"
        #         return best_ann

        # Having this dynamic like above is literally 7x slower, resulting
        # in bottlenecks. Gotta do it the ugly way unfortunately
        ann = self._get_best_ann_by_local_pref(current_ann, new_ann)
        if ann:
            return ann
        else:
            ann = self._get_best_ann_by_as_path(current_ann, new_ann)
            if ann:
                return ann
            else:
                return self._get_best_ann_by_lowest_neighbor_asn_tiebreaker(
                    current_ann, new_ann
                )
        raise GaoRexfordError("No ann was chosen")


def _get_best_ann_by_local_pref(
    self: "BGP", current_ann: Ann, new_ann: Ann
) -> Ann | None:
    """Returns best announcement by local pref, or None if tie"""

    if current_ann.recv_relationship.value > new_ann.recv_relationship.value:
        return current_ann
    elif current_ann.recv_relationship.value < new_ann.recv_relationship.value:
        return new_ann
    else:
        return None


def _get_best_ann_by_as_path(self: "BGP", current_ann: Ann, new_ann: Ann) -> Ann | None:
    """Returns best announcement by as path length, or None if tie

    Shorter AS Paths are better
    """

    if len(current_ann.as_path) < len(new_ann.as_path):
        return current_ann
    elif len(current_ann.as_path) > len(new_ann.as_path):
        return new_ann
    else:
        return None


def _get_best_ann_by_lowest_neighbor_asn_tiebreaker(
    self: "BGP", current_ann: Ann, new_ann: Ann
) -> Ann:
    """Determines if the new ann > current ann by Gao Rexford for ties

    This breaks ties by lowest asn of the neighbor sending the announcement
    So if the two announcements are from the same nieghbor, return current ann
    """

    current_neighbor_asn = current_ann.as_path[min(len(current_ann.as_path), 1)]
    new_neighbor_asn = new_ann.as_path[min(len(new_ann.as_path), 1)]

    if current_neighbor_asn <= new_neighbor_asn:
        return current_ann
    else:
        return new_ann



from .base import (
    AS,
    ASGraph,
    ASGraphCollector,
    ASGraphInfo,
    CustomerProviderLink,
    Link,
    PeerLink,
)
from .caida_as_graph import CAIDAASGraph, CAIDAASGraphCollector, CAIDAASGraphConstructor

__all__ = [
    "ASGraph",
    "AS",
    "ASGraphCollector",
    "ASGraphInfo",
    "CustomerProviderLink",
    "Link",
    "PeerLink",
    "CAIDAASGraphCollector",
    "CAIDAASGraphConstructor",
    "CAIDAASGraph",
]



import logging
from datetime import date
from pathlib import Path

from platformdirs import PlatformDirs, PlatformDirsABC

# NOTE: Can't use getpass here due to windows bug (https://bugs.python.org/issue32731)
DIRS: PlatformDirsABC = PlatformDirs("bgpy", Path.home().name)

SINGLE_DAY_CACHE_DIR: Path = Path(DIRS.user_cache_dir) / str(date.today())
SINGLE_DAY_CACHE_DIR.mkdir(exist_ok=True, parents=True)

bgpy_logger = logging.getLogger("bgpy")
bgpy_logger.setLevel(logging.INFO)



from typing import TYPE_CHECKING

from .ann_container import AnnContainer

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann


class RecvQueue(AnnContainer[str, list["Ann"]]):
    """Adj_RIBs_In for a BGP AS

    Map prefixes to anns sent
    {prefix: list_of_ann}
    """

    def add_ann(self, ann: "Ann"):
        """Appends ann to the list of recieved ann for that prefix

        We don't use defaultdict here because those are not yamlable
        """

        prefix = ann.prefix
        dct = self.data
        list_of_anns = dct.get(prefix)
        if list_of_anns:
            list_of_anns.append(ann)
        else:
            dct[prefix] = [ann]

        # This implementation is much slower
        # self.data[ann.prefix] = self.data.get(ann.prefix, list()) + [ann]

    def get_ann_list(self, prefix: str) -> list["Ann"]:
        """Returns recevied ann list for a given prefix"""

        return self.data.get(prefix, list())



from pathlib import Path

from PIL import Image, ImageFile


class DiagramAggregator:
    """Aggregates diagrams"""

    def __init__(self, base_dir: Path):
        # Needed to aggregate all diagrams
        self.base_dir: Path = base_dir
        self.base_dir.mkdir(parents=True, exist_ok=True)
        # Fix to allow loading of truncated images
        # https://stackoverflow.com/questions/60584155/oserror-image-file-is-truncated
        ImageFile.LOAD_TRUNCATED_IMAGES = True

    def aggregate_diagrams(self):
        """Aggregates all test diagrams for readability into a PDF"""

        # Because we have too many tests, we need to aggregate them for
        # readability
        # https://stackoverflow.com/a/47283224/8903959
        images = [Image.open(x) for x in self.image_paths]
        assert images, "No images were present"
        converted_images = list()
        for img in images:
            if img.mode == "RGBA":
                converted_images.append(img.convert("RGB"))
                img.close()
            else:
                converted_images.append(img)

        # Aggregate all images into a PDF
        converted_images[0].save(
            self.aggregated_diagrams_path,
            "PDF",
            resolution=100.0,
            save_all=True,
            append_images=converted_images[1:],
        )
        for img in converted_images:  # type: ignore
            img.close()

    @property
    def aggregated_diagrams_path(self) -> Path:
        """Returns the path to the aggregated diagrams PDF"""

        return self.base_dir / "aggregated_diagrams.pdf"

    @property
    def image_paths(self) -> list[Path]:
        """Returns paths as strings for all images"""

        return sorted(self.base_dir.glob("*/*png"))



from frozendict import frozendict

from bgpy.as_graphs import ASGraphInfo
from bgpy.as_graphs.base.links import CustomerProviderLink as CPLink
from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import ASPA, BGP
from bgpy.simulation_framework import ForgedOriginPrefixHijack, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

as_graph_info = ASGraphInfo(
    customer_provider_links=frozenset(
        [
            CPLink(provider_asn=1, customer_asn=ASNs.ATTACKER.value),
            CPLink(provider_asn=2, customer_asn=ASNs.ATTACKER.value),
            CPLink(provider_asn=3, customer_asn=ASNs.VICTIM.value),
            CPLink(provider_asn=4, customer_asn=2),
            CPLink(provider_asn=5, customer_asn=4),
            CPLink(provider_asn=5, customer_asn=1),
        ]
    ),
)


desc = "ASPA weirdness. rejects upstream, accepts downstream"

ex_config_029 = EngineTestConfig(
    name="ex_029_aspa_weirdness",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=ForgedOriginPrefixHijack,
        BasePolicyCls=BGP,
        AdoptPolicyCls=ASPA,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                4: ASPA,
                ASNs.VICTIM.value: ASPA,
            }
        ),
    ),
    as_graph_info=as_graph_info,
)



from typing import TYPE_CHECKING, Iterator

from bgpy.shared.enums import Relationships, Timestamps
from bgpy.simulation_engine.policies.rov import ROV

if TYPE_CHECKING:
    from bgpy.as_graphs import AS
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_framework import Scenario


class ROVPPV1Lite(ROV):
    """An Policy that deploys ROV++V1 Lite as defined in the ROV++ paper

    ROV++ Improved Deployable Defense against BGP Hijacking
    """

    name: str = "ROV++V1 Lite"

    def _policy_propagate(
        self,
        neighbor: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> bool:
        """Only propagate announcements that aren't blackholes"""

        # Policy handled this ann for propagation (and did nothing if blackhole)
        return ann.rovpp_blackhole

    def process_incoming_anns(
        self,
        *,
        from_rel: Relationships,
        propagation_round: int,
        scenario: "Scenario",
        reset_q: bool = True,
    ) -> None:
        """Processes all incoming announcements from a specific rel"""

        super().process_incoming_anns(
            from_rel=from_rel,
            propagation_round=propagation_round,
            scenario=scenario,
            reset_q=False,
        )
        self._add_blackholes(from_rel, scenario)

        # It's possible that we had a previously valid prefix
        # Then later recieved a subprefix that was invalid
        # So we must recount the holes of each ann in local RIB
        self._recount_holes(propagation_round)

        self._reset_q(reset_q)

    def _add_blackholes(self, from_rel: "Relationships", scenario: "Scenario") -> None:
        """Adds blackhole announcements to the local RIB

        First add all non routed prefixes from ROAs as blackholes
        Then for each ann in the local RIB, if you received an
        invalid subprefix from the same neighbor,
        add it to the local RIB as a blackhole
        """

        non_routed_blackholes = self._get_non_routed_blackholes_to_add(scenario)
        routed_blackholes = self._get_routed_blackholes_to_add(from_rel, scenario)
        self._add_blackholes_tolocal_rib(non_routed_blackholes + routed_blackholes)

    def _get_non_routed_blackholes_to_add(
        self, scenario: "Scenario"
    ) -> tuple["Ann", ...]:
        """Get all the bholes for non routed prefixes to prevent superprefix attacks"""

        non_routed_blackholes_to_add = list()
        for roa in scenario.roas:
            # ROA is non routed
            if not roa.is_non_routed:
                blackhole_ann = scenario.scenario_config.AnnCls(
                    prefix=str(roa.prefix),
                    next_hop_asn=self.as_.asn,
                    as_path=(self.as_.asn,),
                    timestamp=Timestamps.VICTIM.value,
                    seed_asn=None,
                    roa_valid_length=True,
                    roa_origin=roa.origin,
                    recv_relationship=Relationships.ORIGIN,
                    rovpp_blackhole=True,
                )
                non_routed_blackholes_to_add.append(blackhole_ann)
        return tuple(non_routed_blackholes_to_add)

    def _get_routed_blackholes_to_add(
        self, from_rel: "Relationships", scenario: "Scenario"
    ) -> tuple["Ann", ...]:
        """Get all the bholes from the anns you just recieved"""

        blackholes_to_add = list()
        # Then add blackholes for anns in local RIB when you've
        # recieved an invalid subprefix from the same neighbor
        for _, ann in self.local_rib.items():
            for sub_ann in self._invalid_subprefixes_from_same_neighbor(scenario, ann):
                blackhole = self._copy_and_process(
                    sub_ann,
                    from_rel,
                    overwrite_default_kwargs={
                        "next_hop_asn": self.as_.asn,
                        "rovpp_blackhole": True,
                    },
                )
                blackholes_to_add.append(blackhole)
        return tuple(blackholes_to_add)

    def _invalid_subprefixes_from_same_neighbor(
        self, scenario: "Scenario", ann: "Ann"
    ) -> Iterator["Ann"]:
        """Returns all invalid subprefixes of announcement from the same neighbor"""

        # If we are the origin, then there are zero invalid anns from the same neighbor
        if ann.recv_relationship == Relationships.ORIGIN:
            return ()  # type: ignore
        # For each subprefix in this scenario of the prefix within the local RIB
        for subprefix in scenario.ordered_prefix_subprefix_dict[ann.prefix]:
            # For each subprefix ann that was recieved
            # NOTE: these wouldn't be in the local RIB since they're invalid
            # and dropped by default (but they are recieved so we can check there)
            for sub_ann in self.recv_q.get_ann_list(subprefix):
                # Holes are only from same neighbor
                if (
                    sub_ann.invalid_by_roa
                    # Check the first one in the path since it's already processed
                    and sub_ann.as_path[0] == ann.as_path[1]
                ):
                    yield sub_ann

    def _add_blackholes_tolocal_rib(self, blackholes: tuple["Ann", ...]) -> None:
        """Adds all blackholes to the local RIB"""

        for blackhole in blackholes:
            existing_ann = self.local_rib.get(blackhole.prefix)
            # Don't overwrite valid existing announcements
            if existing_ann is None or existing_ann.invalid_by_roa:
                self.local_rib.add_ann(blackhole)

    def _recount_holes(self, propagation_round: int) -> None:
        # It's possible that we had a previously valid prefix
        # Then later recieved a subprefix that was invalid
        # Or there was previously an invalid subprefix
        # But later that invalid subprefix was removed
        # So we must recount the holes of each ann in local RIB
        # NOTE June 22 2024: I think doing this may require the use of the RIBsIn
        # because you could recieve a subprefix hijack round 1, and round 2 receive
        # the valid prefix from the same neighbor
        # not going to implement because of that, and because I don't think there's
        # a need to, because as far as I know there aren't any two round attacks
        # against ROV++. If someone comes up with one let me know and I can try to
        # help out, email at jfuruness@gmail.com.
        # NOTE: Additionally, we don't account for withdrawals at all...
        if propagation_round != 0:
            raise NotImplementedError("TODO: support ROV++ for multiple rounds")



from yaml import SafeLoader


# https://stackoverflow.com/a/39554610/8903959
class SimulatorLoader(SafeLoader):
    def construct_python_tuple(self, node):
        return tuple(self.construct_sequence(node))


SimulatorLoader.add_constructor(
    "tag:yaml.org,2002:python/tuple", SimulatorLoader.construct_python_tuple
)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Iterator

from yamlable import YamlAble, yaml_info

from bgpy.simulation_engine import Announcement as Ann

from .ann_container import AnnContainer

if TYPE_CHECKING:
    from bgpy.as_graphs import AS


@yaml_info(yaml_tag="SendInfo")
@dataclass
class SendInfo(YamlAble):
    withdrawal_ann: Ann | None = None
    ann: Ann | None = None

    @property
    def anns(self):
        return [x for x in [self.withdrawal_ann, self.ann] if x is not None]

    def __str__(self):
        return f"send_info: ann: {self.ann}, withdrawal {self.withdrawal_ann}"


class SendQueue(AnnContainer[int, dict[str, SendInfo]]):
    """Announcements to be sent for a BGP AS

    {neighbor: {prefix: SendInfo}}
    """

    def add_ann(self, neighbor_asn: int, ann: Ann):
        """Adds Ann to be sent"""

        # Used to be done by the defaultdict
        if neighbor_asn not in self.data:
            self.data[neighbor_asn] = {ann.prefix: SendInfo()}
        if ann.prefix not in self.data[neighbor_asn]:
            self.data[neighbor_asn][ann.prefix] = SendInfo()

        send_info = self.data[neighbor_asn][ann.prefix]

        # If the announcement is a withdraw
        if ann.withdraw:
            # Ensure withdrawls aren't replaced
            msg: str = f"replacing withdrawal? {send_info.withdrawal_ann}"
            assert send_info.withdrawal_ann is None, msg

            # If the withdrawal is equal to ann, delete both
            if send_info.ann is not None and send_info.ann.prefix_path_attributes_eq(
                ann
            ):
                del self.data[neighbor_asn][ann.prefix]
            # If withdrawl is not equal to Ann, add withdrawal
            else:
                send_info.withdrawal_ann = ann

        # If the announcement is not a withdrawal
        else:
            # Never replace valid Ann without a withdrawal
            assert send_info.ann is None, "Replacing valid ann?"
            err = "Can't send identical withdrawal and ann"
            err += f" {ann}, {send_info.withdrawal_ann}"
            assert not ann.prefix_path_attributes_eq(send_info.withdrawal_ann), err

            # Add announcement
            send_info.ann = ann

    def get_send_info(self, neighbor_obj: "AS", prefix: str) -> Ann | None:
        """Returns the SendInfo for a neighbor AS and prefix"""

        return self.data.get(neighbor_obj.asn, dict()).get(prefix)

    def info(self, neighbors: list["AS"]) -> Iterator[tuple["AS", str, Ann]]:
        """Returns neighbor obj, prefix, announcement"""

        for neighbor_obj in neighbors:
            # assert isinstance(neighbor_obj, bgp_as.BGPAS)
            for prefix, send_info in self.data.get(neighbor_obj.asn, dict()).items():
                for ann in send_info.anns:
                    yield neighbor_obj, prefix, ann



from abc import ABC, abstractmethod
from typing import Any


class Link(ABC):
    """Contains a relationship link in a BGP topology"""

    def __init__(self, asn1: int, asn2: int) -> None:
        # Make sure we have asns
        # Make sure the asns is a tuple
        assert isinstance(self.asns, tuple)
        # Make sure the asns is sorted
        assert tuple(sorted(self.asns)) == self.asns

    def __hash__(self) -> int:
        """Hashes used in sets"""

        return hash(self.asns)

    @abstractmethod
    def __eq__(self, other) -> bool:
        raise NotImplementedError

    def __lt__(self, other: Any):
        if isinstance(other, Link):
            return self.__hash__() < other.__hash__()
        else:
            return NotImplemented

    @property
    @abstractmethod
    def asns(self) -> tuple[int, ...]:
        raise NotImplementedError



"""Functions to create ranks for propagation"""

from .base_as import AS


def _assign_propagation_ranks(self):
    """Assigns propagation ranks from the leafs to input_clique"""

    for as_obj in self:
        self._assign_ranks_helper(as_obj, 0)


def _assign_ranks_helper(self, as_obj: AS, rank: int):
    """Assigns ranks to all ases in customer/provider chain recursively"""

    if as_obj.propagation_rank is None or as_obj.propagation_rank < rank:
        as_obj.propagation_rank = rank
        # Only update it's providers if it's rank becomes higher
        # This avoids a double for loop of writes
        for provider_obj in as_obj.providers:
            self._assign_ranks_helper(provider_obj, rank + 1)


def _get_propagation_ranks(self) -> tuple[tuple[AS, ...], ...]:
    """Orders ASes by rank"""

    max_rank: int = max(x.propagation_rank for x in self)
    # Create a list of empty lists
    # Ignore types here for speed purposes
    ranks: list[list[AS]] = [list() for _ in range(max_rank + 1)]
    # Append the ASes into their proper rank
    for as_obj in self:
        ranks[as_obj.propagation_rank].append(as_obj)

    # Create tuple ranks
    return tuple([tuple(sorted(rank)) for rank in ranks])



import pytest

from bgpy.as_graphs import CAIDAASGraphConstructor
from bgpy.simulation_engine import SimulationEngine


@pytest.fixture(scope="session")
def engine() -> SimulationEngine:
    # Engine is not picklable or dillable AT ALL, so do it here
    # (after the multiprocess process has started)
    # Changing recursion depth does nothing
    # Making nothing a reference does nothing
    as_graph = CAIDAASGraphConstructor().run()
    return SimulationEngine(as_graph)



from .graph_factory import GraphFactory
from .line_data import LineData
from .line_info import LineInfo
from .line_properties_generator import LinePropertiesGenerator

__all__ = ["GraphFactory", "LineData", "LineInfo", "LinePropertiesGenerator"]



from typing import Any, Callable
from weakref import proxy

from frozendict import frozendict
from yamlable import YamlAble, yaml_info, yaml_info_decorate

import bgpy
from bgpy.as_graphs.base.as_graph_info import ASGraphInfo
from bgpy.shared.enums import ASGroups, Relationships

from .base_as import AS
from .cone_funcs import (
    _get_and_store_customer_cone_and_set_size,
    _get_and_store_provider_cone_and_set_size,
    _get_as_rank,
    _get_cone_helper,
    _get_size_of_and_store_cone,
)

# can't import into class due to mypy issue
# https://github.com/python/mypy/issues/7045
# Graph building functionality
from .graph_building_funcs import (
    _add_relationships,
    _gen_graph,
    _make_relationships_tuples,
)

# propagation rank building funcs
from .propagation_rank_funcs import (
    _assign_propagation_ranks,
    _assign_ranks_helper,
    _get_propagation_ranks,
)


@yaml_info(yaml_tag="ASGraph")
class ASGraph(YamlAble):
    """BGP Topology"""

    # Graph building functionality
    _gen_graph = _gen_graph
    _add_relationships = _add_relationships
    _make_relationships_tuples = _make_relationships_tuples

    # propagation rank building funcs
    _assign_propagation_ranks = _assign_propagation_ranks
    _assign_ranks_helper = _assign_ranks_helper
    _get_propagation_ranks = _get_propagation_ranks

    # Cone funcs
    _get_size_of_and_store_cone = _get_size_of_and_store_cone
    _get_and_store_customer_cone_and_set_size = (
        _get_and_store_customer_cone_and_set_size
    )
    _get_and_store_provider_cone_and_set_size = (
        _get_and_store_provider_cone_and_set_size
    )
    _get_and_store_provider_cone_and_set_size = (
        _get_and_store_provider_cone_and_set_size
    )
    _get_cone_helper = _get_cone_helper
    _get_as_rank = _get_as_rank

    def __init_subclass__(cls, *args, **kwargs):
        """This method essentially creates a list of all subclasses
        This is allows us to easily assign yaml tags
        """

        super().__init_subclass__(*args, **kwargs)
        # Fix this later once the system test framework is updated
        yaml_info_decorate(cls, yaml_tag=cls.__name__)

    def __init__(
        self,
        as_graph_info: "ASGraphInfo",
        BaseASCls: type[AS] = AS,
        BasePolicyCls: type[bgpy.simulation_engine.Policy] = bgpy.simulation_engine.BGP,
        store_customer_cone_size: bool = True,
        store_customer_cone_asns: bool = False,
        store_provider_cone_size: bool = False,
        store_provider_cone_asns: bool = False,
        yaml_as_dict: frozendict[int, AS] | None = None,
        yaml_ixp_asns: frozenset[int] = frozenset(),
        # Users can pass in any additional AS groups they want to keep track of
        additional_as_group_filters: frozendict[
            str, Callable[["ASGraph"], frozenset[AS]]
        ] = frozendict(),
    ):
        """Reads in relationship data from a TSV and generate graph"""

        if yaml_as_dict is not None:
            # We are coming from YAML, so init from YAML (for testing)
            self._set_yaml_attrs(yaml_as_dict, yaml_ixp_asns)
        else:
            # init as normal, through the as_graph_info
            self._set_non_yaml_attrs(
                as_graph_info,
                BaseASCls,
                BasePolicyCls,
                store_customer_cone_size,
                store_customer_cone_asns,
                store_provider_cone_size,
                store_provider_cone_asns,
            )
        # Set the AS and ASN group groups
        self._set_as_groups(additional_as_group_filters)

    def __eq__(self, other) -> bool:
        if isinstance(other, ASGraph):
            return self.__to_yaml_dict__() == other.__to_yaml_dict__()
        else:
            return NotImplemented

    ##############
    # Init funcs #
    ##############

    def _set_yaml_attrs(
        self,
        yaml_as_dict: frozendict[int, AS],
        yaml_ixp_asns: frozenset[int],
    ) -> None:
        """Generates the AS Graph from YAML"""

        self.ixp_asns: frozenset[int] = yaml_ixp_asns
        self.as_dict: frozendict[int, AS] = yaml_as_dict
        # Convert ASNs to refs
        for as_obj in self.as_dict.values():
            as_obj.as_graph = proxy(self)
            as_obj.peers = tuple([self.as_dict[asn] for asn in as_obj.peers])
            as_obj.customers = tuple([self.as_dict[asn] for asn in as_obj.customers])
            as_obj.providers = tuple([self.as_dict[asn] for asn in as_obj.providers])

        # Used for iteration
        self.ases: tuple[AS, ...] = tuple(self.as_dict.values())
        self.propagation_ranks: tuple[tuple[AS, ...], ...] = (
            self._get_propagation_ranks()
        )

    def _set_non_yaml_attrs(
        self,
        as_graph_info: ASGraphInfo,
        BaseASCls: type["AS"],
        BasePolicyCls: type["bgpy.simulation_engine.Policy"],
        store_customer_cone_size: bool,
        store_customer_cone_asns: bool,
        store_provider_cone_size: bool,
        store_provider_cone_asns: bool,
    ) -> None:
        """Generates the AS graph normally (not from YAML)"""

        assert as_graph_info.ixp_asns is not None
        self.ixp_asns = as_graph_info.ixp_asns
        # Probably there is a better way to do this, but for now we
        # store this as a dict then later make frozendict, thus the type ignore
        self.as_dict: frozendict[int, AS] = dict()  # type: ignore
        # Just adds all ASes to the dict, and adds ixp/input_clique info
        self._gen_graph(
            as_graph_info,
            BaseASCls,
            BasePolicyCls,
        )
        # Can't allow modification of the AS dict since other things like
        # as group filters will be broken then
        self.as_dict = frozendict(self.as_dict)
        # Adds references to all relationships
        self._add_relationships(as_graph_info)
        # Used for iteration
        self.ases = tuple(self.as_dict.values())
        # Remove duplicates from relationships and sort
        self._make_relationships_tuples()
        # Assign propagation rank to each AS
        self._assign_propagation_ranks()
        # Get the ranks for the graph
        self.propagation_ranks = self._get_propagation_ranks()
        # We don't run this every time since it has the runtime greater than the
        # entire graph generation
        if any([store_customer_cone_size, store_customer_cone_asns]):
            # Determine customer cones of all ases
            self._get_size_of_and_store_cone(
                rel_attr=Relationships.CUSTOMERS.name.lower(),
                store_cone_asns=store_customer_cone_asns,
            )
            self._get_as_rank()

        if any([store_provider_cone_size, store_provider_cone_asns]):
            self._get_size_of_and_store_cone(
                rel_attr=Relationships.PROVIDERS.name.lower(),
                store_cone_asns=store_provider_cone_asns,
            )

    def _set_as_groups(
        self,
        additional_as_group_filters: (
            frozendict[str, Callable[["ASGraph"], frozenset[AS]]] | None
        ),
    ) -> None:
        """Sets the AS Groups"""

        as_group_filters: dict[str, Callable[[ASGraph], frozenset[AS]]] = dict(
            self._default_as_group_filters
        )

        if additional_as_group_filters:
            as_group_filters.update(additional_as_group_filters)

        self.as_group_filters: frozendict[str, Callable[[ASGraph], frozenset[AS]]] = (
            frozendict(as_group_filters)
        )

        # Some helpful sets of ases for faster loops
        as_groups: dict[str, frozenset[AS]] = dict()
        asn_groups: dict[str, frozenset[int]] = dict()

        for as_group_key, filter_func in self.as_group_filters.items():
            as_groups[as_group_key] = filter_func(self)
            asn_groups[as_group_key] = frozenset(x.asn for x in filter_func(self))

        # Turn these into frozen dicts. They shouldn't be modified
        self.as_groups: frozendict[str, frozenset[AS]] = frozendict(as_groups)
        self.asn_groups: frozendict[str, frozenset[int]] = frozendict(asn_groups)

    @property
    def _default_as_group_filters(
        self,
    ) -> frozendict[str, Callable[["ASGraph"], frozenset[AS]]]:
        """Returns the default filter functions for AS groups"""

        def ixp_filter(as_graph: "ASGraph") -> frozenset[AS]:
            return frozenset(x for x in as_graph if x.ixp)

        def stub_no_ixp_filter(as_graph: "ASGraph") -> frozenset[AS]:
            return frozenset(x for x in as_graph if x.stub and not x.ixp)

        def multihomed_no_ixp_filter(as_graph: "ASGraph") -> frozenset[AS]:
            return frozenset(x for x in as_graph if x.multihomed and not x.ixp)

        def stubs_or_multihomed_no_ixp_filter(as_graph: "ASGraph") -> frozenset[AS]:
            return frozenset(
                x for x in as_graph if (x.stub or x.multihomed) and not x.ixp
            )

        def input_clique_no_ixp_filter(as_graph: "ASGraph") -> frozenset[AS]:
            return frozenset(x for x in as_graph if x.input_clique and not x.ixp)

        def etc_no_ixp_filter(as_graph: "ASGraph") -> frozenset[AS]:
            return frozenset(
                x
                for x in as_graph
                if not (x.stub or x.multihomed or x.input_clique or x.ixp)
            )

        def transit_no_ixp_filter(as_graph: "ASGraph") -> frozenset[AS]:
            return frozenset(x for x in as_graph if x.transit and not x.ixp)

        def all_no_ixp_filter(as_graph: "ASGraph") -> frozenset[AS]:
            return frozenset(list(as_graph))

        return frozendict(
            {
                ASGroups.IXPS.value: ixp_filter,
                ASGroups.STUBS.value: stub_no_ixp_filter,
                ASGroups.MULTIHOMED.value: multihomed_no_ixp_filter,
                ASGroups.STUBS_OR_MH.value: stubs_or_multihomed_no_ixp_filter,
                ASGroups.INPUT_CLIQUE.value: input_clique_no_ixp_filter,
                ASGroups.ETC.value: etc_no_ixp_filter,
                ASGroups.TRANSIT.value: transit_no_ixp_filter,
                ASGroups.ALL_WOUT_IXPS.value: all_no_ixp_filter,
            }
        )

    ##############
    # Yaml funcs #
    ##############

    def __to_yaml_dict__(self) -> dict[str, Any]:
        """Optional method called when yaml.dump is called"""

        return {
            "as_dict": dict(self.as_dict.items()),
            "ixp_asns": list(self.ixp_asns),
        }

    @classmethod
    def __from_yaml_dict__(cls, dct, yaml_tag) -> Any:
        """Optional method called when yaml.load is called"""

        return cls(
            ASGraphInfo(),
            yaml_as_dict=frozendict(dct["as_dict"]),
            yaml_ixp_asns=frozenset(dct["ixp_asns"]),
        )

    ##################
    # Iterator funcs #
    ##################

    # https://stackoverflow.com/a/7542261/8903959
    def __getitem__(self, index: int) -> AS:
        return self.ases[index]

    def __len__(self) -> int:
        return len(self.as_dict)



from ipaddress import ip_network
from typing import TYPE_CHECKING, Optional

from roa_checker import ROA

from bgpy.shared.enums import Prefixes, Timestamps
from bgpy.simulation_framework.scenarios.scenario import Scenario

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_engine import BaseSimulationEngine


class NonRoutedPrefixHijack(Scenario):
    """Non routed prefix hijack (ROA of AS 0)"""

    def _get_announcements(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns non routed prefix announcement from attacker

        for subclasses of this EngineInput, you can set AnnCls equal to
        something other than Announcement
        """

        anns = list()
        for attacker_asn in self.attacker_asns:
            anns.append(
                self.scenario_config.AnnCls(
                    prefix=Prefixes.PREFIX.value,
                    as_path=(attacker_asn,),
                    timestamp=Timestamps.ATTACKER.value,
                )
            )
        return tuple(anns)

    def _get_roas(
        self,
        *,
        announcements: tuple["Ann", ...] = (),
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple[ROA, ...]:
        """Returns a tuple of ROAs"""

        return (ROA(ip_network(Prefixes.PREFIX.value), 0),)



from .as_graph import AS, ASGraph
from .as_graph_collector import ASGraphCollector
from .as_graph_constructor import ASGraphConstructor
from .as_graph_info import ASGraphInfo
from .links import CustomerProviderLink, Link, PeerLink

__all__ = [
    "ASGraph",
    "AS",
    "ASGraphCollector",
    "ASGraphConstructor",
    "ASGraphInfo",
    "CustomerProviderLink",
    "Link",
    "PeerLink",
]



import csv
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

from frozendict import frozendict

from bgpy.shared.constants import bgpy_logger

if TYPE_CHECKING:
    from .as_graph import ASGraph
    from .as_graph_collector import ASGraphCollector
    from .as_graph_info import ASGraphInfo


class ASGraphConstructor(ABC):
    def __init__(
        self,
        ASGraphCollectorCls: type["ASGraphCollector"],
        ASGraphCls: type["ASGraph"],
        as_graph_collector_kwargs=frozendict(),
        as_graph_kwargs=frozendict(),
        tsv_path: Path | None = None,
        stubs: bool = True,
    ) -> None:
        """Stores download time and cache_dir instance vars and creates dir"""

        self.as_graph_collector: ASGraphCollector = ASGraphCollectorCls(
            **as_graph_collector_kwargs
        )
        self.ASGraphCls: type[ASGraph] = ASGraphCls
        self.as_graph_kwargs = as_graph_kwargs
        self.tsv_path: Path | None = tsv_path
        self.stubs: bool = stubs

    def run(self) -> "ASGraph":
        """Generates AS graph in the following steps:

        1. download file from source using the GraphCollector
        2. parse downloaded file to get ASGraphInfo object
        3. Generate the graph based on ASGraphInfo object
        4. Write to tsv_path if it is set
        5. Return ASGraph
        """

        # Download file (for ex: from CAIDA)
        dl_path = self.as_graph_collector.run()
        # Get ASGraphInfo from downloaded file
        as_graph_info = self._get_as_graph_info(dl_path)
        # Generate AS Graph from ASGraphInfo
        as_graph = self._get_as_graph(as_graph_info)
        if not self.stubs:
            invalid_asns = frozenset([as_obj.asn for as_obj in as_graph if as_obj.stub])
            # Get ASGraphInfo from downloaded file
            as_graph_info = self._get_as_graph_info(dl_path, invalid_asns)
            # Generate AS Graph from ASGraphInfo
            as_graph = self._get_as_graph(as_graph_info)

        # Write to TSV if tsv_path is set
        self.write_tsv(as_graph, self.tsv_path)
        return as_graph

    def remove_stubs(self, as_graph: "ASGraph") -> None:
        """Removes stubs from as graph"""

        stubs = [as_obj for as_obj in as_graph if as_obj.stub]
        for stub in stubs:
            for provider in stub.providers:
                customers = tuple([x for x in provider.customers if x.asn != stub.asn])
                provider.customers = customers

    @staticmethod
    def write_tsv(as_graph: "ASGraph", tsv_path: Path | None = None) -> None:
        """Writes AS Graph to TSV"""

        if tsv_path:
            bgpy_logger.info(
                f"Writing as graph to {tsv_path} "
                "if you want to save time and not do this, pass tsv_path=None "
                "to the run function"
            )

            with tsv_path.open(mode="w") as f:
                # Get columns
                cols: list[str] = list(
                    next(iter(as_graph.as_dict.values())).db_row.keys()
                )
                writer = csv.DictWriter(f, fieldnames=cols, delimiter="\t")
                writer.writeheader()
                for x in as_graph.as_dict.values():
                    writer.writerow(x.db_row)

    @abstractmethod
    def _get_as_graph_info(
        self, dl_path: Path, invalid_asns: frozenset[int] = frozenset()
    ) -> "ASGraphInfo":
        """Reads cached directory, gets AS graph info and returns ASGraphInfo

        dl_path is the Path to the downloaded file
        """
        raise NotImplementedError

    @abstractmethod
    def _get_as_graph(self, as_graph_info: "ASGraphInfo") -> "ASGraph":
        """Returns AS Graph based on ASGraphInfo"""
        raise NotImplementedError



from bgpy.simulation_engine.policies.rov.rov_full import ROVFull

from .bgpisec import BGPiSec


class BGPiSecFull(BGPiSec, ROVFull):
    """BGP-iSec with withdrawals, ribs in and out"""

    name = "BGP-iSec Full"



from .data_point_agg_data import DataPointAggData
from .data_point_key import DataPointKey
from .graph_category import GraphCategory
from .graph_data_aggregator import GraphDataAggregator
from .trial_data import TrialData

__all__ = [
    "DataPointAggData",
    "DataPointKey",
    "GraphCategory",
    "GraphDataAggregator",
    "TrialData",
]



from bgpy.as_graphs.base import ASGraph


class CAIDAASGraph(ASGraph):
    """For now this is just a copy of the AS Graph

    Since CAIDA is the only AS Graph we currently support
    but that could change in the future
    """

    pass



# YAML STUFF
from collections.abc import Iterable
from typing import Any

import yaml
from yamlable import YamlCodec

from bgpy.shared.constants import bgpy_logger
from bgpy.shared.enums import YamlAbleEnum
from bgpy.simulation_engine.ann_containers.ann_container import AnnContainer

from .simulator_loader import SimulatorLoader

# 2-way mappings between the types and the yaml tags
types_to_yaml_tags: dict[type[Any], str] = {
    X: X.yaml_suffix() for X in YamlAbleEnum.yamlable_enums()
}
types_to_yaml_tags.update({Cls: Cls.__name__ for Cls in AnnContainer.subclasses})

yaml_tags_to_types = {v: k for k, v in types_to_yaml_tags.items()}


class SimulatorCodec(YamlCodec):
    @classmethod
    def get_yaml_prefix(cls):
        """Root yaml tag"""

        return "!simulator_codec/"

    @classmethod
    def get_known_types(cls) -> Iterable[type[Any]]:
        """return the list of types that we know how to encode"""

        return types_to_yaml_tags.keys()

    @classmethod
    def is_yaml_tag_supported(cls, yaml_tag_suffix: str) -> bool:
        """return True if the given yaml tag suffix is supported"""

        return yaml_tag_suffix in yaml_tags_to_types

    @classmethod
    def from_yaml_dict(cls, yaml_tag_suffix: str, dct, **kwargs):
        """Creates object related to the given tag, from decoded dict"""

        typ = yaml_tags_to_types[yaml_tag_suffix]
        try:
            if issubclass(typ, YamlAbleEnum):
                # Don't use unessecary name
                return typ(value=dct["value"])
            elif issubclass(typ, AnnContainer):
                return typ.__from_yaml_dict__(dct=dct, yaml_tag=yaml_tag_suffix)
            else:
                return typ(**dct)
        except Exception as e:
            # For some reason YamlAble library suppresses these errors...
            bgpy_logger.exception(str(e))
            raise

    @classmethod
    def to_yaml_dict(cls, obj) -> tuple[str, Any]:
        """Converts objects to yaml dicts"""

        if isinstance(obj, YamlAbleEnum):
            return types_to_yaml_tags[type(obj)], {"value": obj.value, "name": obj.name}
        elif isinstance(obj, AnnContainer):
            rv = types_to_yaml_tags[type(obj)], obj.__to_yaml_dict__()
            return rv
        else:
            # Encode the given object and also return the tag it should have
            return types_to_yaml_tags[type(obj)], vars(obj)

    def dump(self, obj, path=None):
        # https://stackoverflow.com/a/30682604/8903959
        # Ignores references for more readable output
        yaml.Dumper.ignore_aliases = lambda *args: True  # type: ignore
        if path is None:
            yaml.dump(obj)
        else:
            with path.open(mode="w") as f:
                yaml.dump(obj, f)

    def load(self, path):
        with path.open(mode="r") as f:
            # This isn't insecure, ignore S506
            return yaml.load(f, Loader=SimulatorLoader)  # noqa: S506


SimulatorCodec.register_with_pyyaml()



from typing import Iterable

from bgpy.simulation_engine import Announcement as Ann

from .ann_container import AnnContainer


class RIBsOut(AnnContainer[int, dict[str, Ann]]):
    """Incomming announcements for a BGP AS

    neighbor: {prefix: announcement}
    """

    def get_ann(self, neighbor_asn: int, prefix: str) -> Ann | None:
        """Returns Ann for a given neighbor asn and prefix"""

        return self.data.get(neighbor_asn, dict()).get(prefix)

    def add_ann(self, neighbor_asn: int, ann: Ann) -> None:
        """Adds announcement to the ribs out"""

        if neighbor_asn in self.data:
            self.data[neighbor_asn][ann.prefix] = ann
        else:
            self.data[neighbor_asn] = {ann.prefix: ann}

    def remove_entry(self, neighbor_asn: int, prefix: str) -> None:
        """Removes ann from ribs out"""

        del self.data[neighbor_asn][prefix]

    def neighbors(self) -> Iterable[int]:
        """Return all neighbors from the ribs out"""

        return self.data.keys()



from typing import TYPE_CHECKING

from .ann_container import AnnContainer

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann


class LocalRIB(AnnContainer[str, "Ann"]):
    """Local RIB for a BGP AS"""

    def add_ann(self, ann: "Ann"):
        """Adds an announcement to local rib with prefix as key"""

        self.data[ann.prefix] = ann



from bgpy.simulation_engine.policies.rov.rov_full import ROVFull

from .bgpisec_transitive_only_to_customers import BGPiSecTransitiveOnlyToCustomers


class BGPiSecTransitiveOnlyToCustomersFull(BGPiSecTransitiveOnlyToCustomers, ROVFull):
    """BGP-iSecTransitiveOTC with withdrawals, ribs in and out"""

    name = "BGP-iSec Transitive + OTC Full"



from bgpy.simulation_engine.policies.rov import ROVFull

from .rovpp_v2_improved_lite import ROVPPV2ImprovedLite


class ROVPPV2ImprovedLiteFull(ROVPPV2ImprovedLite, ROVFull):
    """Full version of ROVPPV2ImprovedLite"""

    name: str = "ROV++V2i Lite Full"



from frozendict import frozendict

from bgpy.as_graphs import ASGraphInfo
from bgpy.as_graphs.base.links import CustomerProviderLink as CPLink
from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import ASPA, BGP
from bgpy.simulation_framework import ForgedOriginPrefixHijack, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

as_graph_info_no_downstream = ASGraphInfo(
    customer_provider_links=frozenset(
        [
            CPLink(provider_asn=1, customer_asn=ASNs.ATTACKER.value),
            CPLink(provider_asn=2, customer_asn=ASNs.ATTACKER.value),
            CPLink(provider_asn=1, customer_asn=2),
            CPLink(provider_asn=3, customer_asn=ASNs.VICTIM.value),
        ]
    ),
)


desc = (
    "Origin hijack against ASPASimple\n"
    "Testing that ASPA rejects from the upstream, but accepts from downstream"
)

ex_config_023 = EngineTestConfig(
    name="ex_023_origin_aspa_simple_downstream_verification",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=ForgedOriginPrefixHijack,
        BasePolicyCls=BGP,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                2: ASPA,
                ASNs.VICTIM.value: ASPA,
            }
        ),
    ),
    as_graph_info=as_graph_info_no_downstream,
)



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import ASPA, BGP
from bgpy.simulation_framework import AccidentalRouteLeak, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = (
    "Route leak to check when v_max_complement==u_min\n"
    " (this is merely to check functionality)"
)

ex_config_028 = EngineTestConfig(
    name="ex_028_route_leak_aspa_simple_u_min_v_max_check",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=AccidentalRouteLeak,
        BasePolicyCls=BGP,
        AdoptPolicyCls=ASPA,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                1: ASPA,
                2: ASPA,
                3: ASPA,
                4: ASPA,
                5: ASPA,
                8: ASPA,
                9: ASPA,
                10: ASPA,
                11: ASPA,
                12: ASPA,
                ASNs.VICTIM.value: ASPA,
                ASNs.ATTACKER.value: ASPA,
            }
        ),
    ),
    as_graph_info=as_graph_info_000,
)
